

from mmisdk.adapters.qredo_create_tx_extra_params import QredoCreateTxExtraParams
from mmisdk.common.custodian import Custodian
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction_params import EIP1559TxParams
from mmisdk.common.transaction_params import LegacyTxParams
from mmisdk.qredo.qredo_client import QredoClient
from mmisdk.qredo.qredo_new_transaction import QredoNewTransaction
from mmisdk.qredo.qredo_transaction_info import QredoTransactionInfo
from mmisdk.utils.map_transaction_status import map_transaction_status


class QredoAdapter(Custodian):

    def __init__(self, client: QredoClient) -> None:
        self.client = client

    def get_transaction(self, transaction_id: str, chain_id=1) -> Transaction:
        qredo_transaction_info = self.client.get_transaction(transaction_id)
        return self.__map_qredo_transaction_info_to_transaction(qredo_transaction_info)

    def create_transaction(self, tx_params, extra_params) -> Transaction:
        # Parse and type check params
        type_ = tx_params["type"] if 'type' in tx_params else 1
        tx_params_parsed = EIP1559TxParams(**tx_params) if type_ == 2 else LegacyTxParams(**tx_params)
        extra_params_parsed = QredoCreateTxExtraParams(**extra_params)

        gas_limit = tx_params_parsed.gas if type_ == 1 else None
        max_priority_fee_per_gas = tx_params_parsed.maxPriorityFeePerGas if type_ == 2 else None
        max_fee_per_gas = tx_params_parsed.maxFeePerGas if type_ == 2 else None

        new_transaction = QredoNewTransaction(
            from_=tx_params_parsed.from_,
            to=tx_params_parsed.to,
            value=tx_params_parsed.value,
            gasPrice=tx_params_parsed.gasPrice,
            maxPriorityFeePerGas=max_priority_fee_per_gas,
            maxFeePerGas=max_fee_per_gas,
            gasLimit=gas_limit,
            data=tx_params_parsed.data or "",
            chainID=extra_params_parsed.chainID
        )
        qredo_transaction_info = self.client.create_transaction(new_transaction)
        return self.__map_qredo_transaction_info_to_transaction(qredo_transaction_info)

    def __map_qredo_transaction_info_to_transaction(self, qredo_transaction_info: QredoTransactionInfo) -> Transaction:
        return Transaction(
            id=qredo_transaction_info.txID,
            type="1" if qredo_transaction_info.gasPrice is not None else "2",
            from_=qredo_transaction_info.from_,
            to=qredo_transaction_info.to,
            value=qredo_transaction_info.value,
            gas=qredo_transaction_info.gasLimit,
            gasPrice=qredo_transaction_info.gasPrice,
            maxPriorityFeePerGas=qredo_transaction_info.maxPriorityFeePerGas,
            maxFeePerGas=qredo_transaction_info.maxFeePerGas,
            nonce=qredo_transaction_info.nonce,
            data=qredo_transaction_info.data if qredo_transaction_info.data != "" else None,
            hash=qredo_transaction_info.txHash if qredo_transaction_info.txHash != "" else None,
            status=map_transaction_status(qredo_transaction_info.status, "Unknown"),
        )



from typing import List

import requests

from mmisdk.cactus.cactus_get_transactions_params import CactusGetTransactionsParams
from mmisdk.cactus.cactus_transaction_detail import CactusTransactionDetail
from mmisdk.cactus.cactus_tx_params import CactusTXParams
from mmisdk.common.abstract_client import AbstractClient
from mmisdk.utils.inject_from_ import inject_from_on_all
from cachetools import TTLCache


class CactusClient(AbstractClient):

    def __init__(self, api_url, token):
        self.api_url = api_url
        self.token = token
        self.cache = TTLCache(maxsize=1, ttl=86400)

    def __create_access_token(self):
        """Internal method that creates an access token, necessary to authenticate other calls against Cactus Custody's API.

        Returns:
            The access token as a string.
        """
        url = f"{self.api_url}/tokens"
        payload = {"grantType": "refresh_token",
                   "refreshToken": self.token}
        headers = {"Content-Type": "application/json"}

        response = requests.post(url,  json=payload, headers=headers)
        response_json = response.json()

        if 'jwt' not in response_json:
            raise requests.HTTPError(
                f"Couldn't create the access token, {response_json}")

        access_token = response_json["jwt"]
        return access_token

    def __create_access_token_cached(self):
        """Lazily fetch the access token, looking up in the cache before. If the access is not in the cache, we fetch it then store it.
        Note that we have to dynamically re-recreate the cache after fetching the token, because the TTL we want to use is extracted from the token's expires_in.
        """
        if 'access_token' not in self.cache:
            access_token = self.__create_access_token()
            self.cache['access_token'] = access_token

        return self.cache['access_token']

    def __get_headers(self):
        """Internal method that creates the HTTP header to use for the calls against Cactus Custody's API. It includes an authorization header that uses the access token.

        Returns:
            The HTTP header as a dictionary.
        """
        access_token = self.__create_access_token_cached()
        return {
            "Content-Type": "application/json",
            "Authorization": 'Bearer ' + access_token
        }

    def get_transactions(self, params: CactusGetTransactionsParams) -> List[CactusTransactionDetail]:
        url = f"{self.api_url}/transactions"
        querystring = params.dict()
        querystring["from"] = querystring["from_"]
        payload = ""
        headers = self.__get_headers()

        response = requests.get(url, data=payload, headers=headers, params=querystring)

        # Parse and validate response
        response_json = response.json()
        response_json_parsed = inject_from_on_all(response_json)
        return list(map(lambda json: CactusTransactionDetail(**json), response_json_parsed))

    def create_transaction(self, chain_id: str, params: CactusTXParams) -> CactusTransactionDetail:
        url = f"{self.api_url}/transactions"
        headers = self.__get_headers()
        querystring = {"chainId": chain_id}
        payload = params.dict()
        payload["from"] = payload["from_"]
        response = requests.post(url, json=payload, headers=headers, params=querystring)
        response_json = response.json()

        if response.status_code != 200:
            raise requests.HTTPError(
                f"Couldn't create the transaction. Request failed with status {response.status_code}: {response_json}")

        # Parse response
        response_json["from_"] = response_json["from"]
        return CactusTransactionDetail(**response_json)



from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class BaseTxParams(BaseModel):
    from_: EthereumAddress = Field(None, alias="from")   # In Python, "from" is a reserved keyword
    to: EthereumAddress
    value: Optional[DecString]
    data: Optional[HexString]
    gas: Optional[DecString]  # Same as gasLimit
    nonce: Optional[DecString]


class LegacyTxParams(BaseTxParams):
    type = 1
    gasPrice: DecString


class EIP1559TxParams(BaseTxParams):
    type = 2
    maxPriorityFeePerGas: DecString
    maxFeePerGas: DecString



MOCK_CACTUS_TRANSACTION_DETAIL_1 = {
    "transactionStatus": "completed",
    "transactionHash": "0xc062e8e89fa6447b59fc981fa6be352d012bd8bf92bb6833626bcb34c89ce455",
    "custodian_transactionId": "VHV3LIBLVND888888000431",
    "gasPrice": "1000000008",
    "maxFeePerGas": "3500000012",
    "maxPriorityFeePerGas": "1000000000",
    "gasLimit": "1500000",
    "nonce": "4",
    "from": "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29",
    "signature": None,
}



import unittest
from unittest import mock

from mmisdk.mocks.mock_qredo_requests_get import mock_qredo_requests_get
from mmisdk.mocks.mock_qredo_requests_post import mock_qredo_requests_post
from mmisdk.qredo.qredo_client import QredoClient
from mmisdk.qredo.qredo_new_transaction import QredoNewTransaction
from mmisdk.qredo.qredo_transaction_info import QredoTransactionStatus


class QredoClientTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.client = QredoClient("http://some-url", "some-refresh-token")

    def test_should_be_defined(self):
        self.assertIsNotNone(self.client)

    @mock.patch('requests.get', side_effect=mock_qredo_requests_get)
    @mock.patch('requests.post', side_effect=mock_qredo_requests_post)
    def test_should_get_transaction(self, mock_get, mock_post):
        """We mock the methods requests.get and requests.post via the above decorators"""

        qredo_transaction_info = self.client.get_transaction(
            "2ELuLA4HWIzPBQB1MSLmbOxtoB1")
        self.assertEqual(qredo_transaction_info.txID, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')
        self.assertEqual(qredo_transaction_info.txHash, '0x6b5dc3519b325be4e6f9512f8d3b1fdb7af22a996e3f6bb1a898a28ef05c1963')
        self.assertEqual(qredo_transaction_info.status, QredoTransactionStatus.AUTHORIZED)
        self.assertEqual(qredo_transaction_info.nonce, 0)
        self.assertEqual(qredo_transaction_info.from_,
                         '0x62468FD916bF27A3b76d3de2e5280e53078e13E1')
        self.assertEqual(
            qredo_transaction_info.to, '0x9999999999999999999999999999999999999999')
        self.assertEqual(qredo_transaction_info.value, '1')
        self.assertEqual(qredo_transaction_info.gasPrice, '1000')
        self.assertEqual(qredo_transaction_info.maxPriorityFeePerGas, None)
        self.assertEqual(qredo_transaction_info.maxFeePerGas, None)
        self.assertEqual(qredo_transaction_info.gasLimit, '21000')
        self.assertEqual(qredo_transaction_info.data, '')
        self.assertEqual(qredo_transaction_info.rawTX, '4YCCA-iCUgiUYkaP2Ra_J6O3bT3i5SgOUweOE-EBgICAgA')
        self.assertEqual(qredo_transaction_info.createdBy, 'EFvSMt9uGTEsDCx22sYB8BPz1bCyxAZNPKDWpJVsLKM9')
        self.assertEqual(qredo_transaction_info.network, '')
        self.assertEqual(qredo_transaction_info.chainID, '1')

    @mock.patch('requests.post', side_effect=mock_qredo_requests_post)
    def test_should_create_transaction(self, mock_post):
        """We mock the methods requests.get and requests.post via the above decorators"""
        new_transaction = QredoNewTransaction(
            from_="0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
            to="0x9999999999999999999999999999999999999999",
            value="1",
            gasPrice="1000",
            maxPriorityFeePerGas=None,
            maxFeePerGas=None,
            gasLimit="21000",
            data="",
            chainID="1",
        )
        transaction = self.client.create_transaction(new_transaction)
        self.assertEqual(transaction.txID, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')

    @mock.patch('requests.post', side_effect=mock_qredo_requests_post)
    def test___create_access_token_cached_should_call_once(self, mock_post):
        access_token_1 = self.client._QredoClient__create_access_token_cached()
        access_token_2 = self.client._QredoClient__create_access_token_cached()
        access_token_3 = self.client._QredoClient__create_access_token_cached()
        access_token_4 = self.client._QredoClient__create_access_token_cached()
        access_token_5 = self.client._QredoClient__create_access_token_cached()
        mock_post.assert_called_once()
        self.assertEqual(access_token_1, access_token_2)
        self.assertEqual(access_token_2, access_token_3)
        self.assertEqual(access_token_3, access_token_4)
        self.assertEqual(access_token_4, access_token_5)


if __name__ == "__main__":
    unittest.main()



from copy import copy
from typing import Dict
from typing import List


def inject_from_(dict: Dict) -> Dict:
    """This method returns a copy of the passed dict, injecting a field "from_", using the same value as field "from". This is needed because 'from' is a reserved keyword in Python."""
    result = copy(dict)
    result["from_"] = result["from"]
    return result


def inject_from_on_all(dict: List[Dict]) -> List[Dict]:
    """"This method returns a copy of the passed list of dicts, injecting a field "from_" on each of them, using the same value as field "from"."""
    return list(map(inject_from_, dict))




from typing import List

import requests
from pydantic import HttpUrl

from mmisdk.mmiconfig.custodian_config import CustodianConfig
from mmisdk.mmiconfig.mmi_config import MMIConfig
from mmisdk.mmiconfig.mmiconfig_legacy_override import MMICONFIG_LEGACY_OVERRIDE

MMI_CONFIGURATION_API_DEV = "https://mmi-configuration-api.codefi.network/v1/configuration/default"
MMI_CONFIGURATION_API_PROD = "https://mmi-configuration-api.mmi-prod.codefi.network/v1/configuration/default"


class MMIConfigClient():

    def __get_mmi_config(self, api_url: HttpUrl) -> MMIConfig:
        response = requests.get(api_url, timeout=1)

        if response.status_code != 200:
            raise requests.HTTPError(
                f'Could not fetch the MMI Configuration in dev. Request failed with code {response.status_code}, {response.text}.')

        return MMIConfig(**response.json())

    def get_merged_custodian_configs(self) -> List[CustodianConfig]:
        """
        Fetch custodian configurations from the MMI configuration API in dev, in prod.
        Then apply the overrides for the legacy custodians.

        Returns the merged list of custodian configs.
        """
        mmi_config_dev = self.__get_mmi_config(MMI_CONFIGURATION_API_DEV)
        mmi_config_prod = self.__get_mmi_config(MMI_CONFIGURATION_API_PROD)
        custodians_config_dev = mmi_config_dev.custodians
        custodians_config_prod = mmi_config_prod.custodians
        custodians_config_legacy = MMICONFIG_LEGACY_OVERRIDE.custodians

        legacy_custodian_names = list(map(lambda config: config.name, custodians_config_legacy))

        # Remove the configs for legacy custodians. We override them with harcoded values
        custodians_config_dev_cleaned = list(filter(lambda config: config.name not in legacy_custodian_names, custodians_config_dev))
        custodians_config_prod_cleaned = list(filter(lambda config: config.name not in legacy_custodian_names, custodians_config_prod))

        return custodians_config_dev_cleaned + custodians_config_prod_cleaned + custodians_config_legacy



from typing import Optional

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString


class CactusCreateTxExtraParams(BaseModel):
    chainId: DecString
    note: Optional[str]



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_CACTUS_DEV"]

# Create the custodian, using the factory
custodian = factory.create_for("cactus-dev", token)

# Get the transaction
transaction = custodian.get_transaction("5CM05NCLMRD888888000800", 5)

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='5CM05NCLMRD888888000800'
# type='1'
# from_='0xFA42B2eCf59abD6d6BD4BF07021D870E2FC0eF20'
# to=None
# value=None
# gas='133997'
# gasPrice='2151'
# maxPriorityFeePerGas=None
# maxFeePerGas=None
# nonce=''
# data=None
# hash=None
# status=TransactionStatus(finished=False, submitted=False, signed=False, success=False, displayText='Created', reason='Unknown')



from mmisdk.mocks.mock_qredo_transaction_info import MOCK_QREDO_TRANSACTION_INFO
from mmisdk.mocks.mock_response import MockResponse


def mock_qredo_requests_post(*args, **kwargs):
    """ This method will be used by the mock to replace requests.post"""

    # Mock the request that creates the access token
    if args[0].__contains__('/connect/token'):
        return MockResponse({"access_token": "some-access-token", "expires_in": 86400}, 200)

    # Mock the request that creates a transaction
    if args[0].__contains__('/connect/transaction'):
        return MockResponse(MOCK_QREDO_TRANSACTION_INFO, 200)

    return MockResponse(None, 404)



from typing import Optional
from typing import Union

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.empty_string import EmptyString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class CactusTransactionDetail(BaseModel):
    """ Refer to ./custody-defi-api-4.yaml for more details """
    transactionStatus: str
    transactionHash: Optional[HexString]
    custodian_transactionId: Optional[str]
    gasPrice: Optional[DecString]
    maxFeePerGas: Optional[DecString]
    maxPriorityFeePerGas: Optional[DecString]
    gasLimit: Optional[DecString]
    nonce: Union[DecString, EmptyString, None]
    from_: Optional[EthereumAddress] = Field(None, alias="from")
    signature: Optional[str]



from typing import Generic
from typing import TypeVar

from pydantic.generics import GenericModel

from mmisdk.bitgo.bitgo_response_meta import BitgoResponseMeta

T = TypeVar("T")


class BitgoResponse(GenericModel, Generic[T]):
    data: T
    _meta: BitgoResponseMeta



from mmisdk.adapters.bitgo_adapter import BitgoAdapter
from mmisdk.adapters.cactus_adapter import CactusAdapter
from mmisdk.adapters.jsonrpc_adapter import JsonRpcAdapter
from mmisdk.adapters.qredo_adapter import QredoAdapter
from mmisdk.bitgo.bitgo_client import BitgoClient
from mmisdk.cactus.cactus_client import CactusClient
from mmisdk.jsonrpc.jsonrpc_client import JsonRpcClient
from mmisdk.qredo.qredo_client import QredoClient

"""
Map each custodian name to its appropriate Client class, and to its appropriate Adapter class.
"""

CUSTODIAN_NAME_TO_CLIENT_ADAPTER = {
    "qredo": {
        "client": QredoClient,
        "adapter": QredoAdapter
    },
    "qredo-dev": {
        "client": QredoClient,
        "adapter": QredoAdapter
    },
    "cactus": {
        "client": CactusClient,
        "adapter": CactusAdapter
    },
    "cactus-dev": {
        "client": CactusClient,
        "adapter": CactusAdapter
    },
    "bitgo": {
        "client": BitgoClient,
        "adapter": BitgoAdapter
    },
    "bitgo-test": {
        "client": BitgoClient,
        "adapter": BitgoAdapter
    },
    "default": {
        "client": JsonRpcClient,
        "adapter": JsonRpcAdapter
    }
}



MOCK_MMICONFIG_DEV = {
    "portfolio": None,
    "custodians": [
        {
            "refreshTokenUrl": "https://hi.ho",
            "name": "saturn",
            "apiBaseUrl": "https://test.net",
        },
    ]
}



import os

from mmisdk.factory.custodian_factory import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_QREDO_DEV"]

custodian = factory.create_for("qredo-dev", token)

# Get the transaction
transaction = custodian.get_transaction("2ELvFICFt3RnXWdyxjkMvFN80xr")

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='2ELvFICFt3RnXWdyxjkMvFN80xr'
# type='1'
# from_='0x62468FD916bF27A3b76d3de2e5280e53078e13E1'
# to='0x62468FD916bF27A3b76d3de2e5280e53078e13E1'
# value='1'
# gas='21000'
# gasPrice='1000'
# maxPriorityFeePerGas=None
# maxFeePerGas=None
# nonce='0'
# data=''
# hash=''
# status=TransactionStatus(finished=False, submitted=False, signed=False, success=False, displayText='Unknown', reason='Unknown')



from typing import Generic
from typing import List
from typing import TypeVar

from pydantic.generics import GenericModel

from mmisdk.bitgo.bitgo_response_meta import BitgoResponseMeta

T = TypeVar("T")


class BitgoListResponse(GenericModel, Generic[T]):
    data: List[T]
    meta: BitgoResponseMeta




from mmisdk.common.custodian import Custodian
from mmisdk.factory.custodian_name_to_client_adapter import CUSTODIAN_NAME_TO_CLIENT_ADAPTER
from mmisdk.mmiconfig.mmiconfig_client import MMIConfigClient


class CustodianFactory:
    """Utility class to instantiate custodians."""

    def __init__(self) -> None:
        mmiconfig_client = MMIConfigClient()
        self.custodians_config = mmiconfig_client.get_merged_custodian_configs()

    def get_supported_custodians(self):
        """Lists the names of custodians supported by the library. Use this name in the method "create_for" to instantiate a custodian object.

        Returns:
            The names of supported custodians, as a list of strings
        """
        return list(map(lambda config: config.name, self.custodians_config))

    def create_for(self, custodian_name: str, token: str) -> Custodian:
        """Creates an custodian instance for the passed custodian name.

        Args:
            custodian_name: The custodian name. Call the method "get_supported_custodians" to check the avaialble values.
            token: A token provided to you by the custodian. Depending on the custodian, it could be either a refresh token
            or an access token. Regardless of which, the library takes care of the authentication flow.

        Returns:
            The custodian instance.
        """
        configs_with_name = list(filter(
            lambda config: config.name == custodian_name, self.custodians_config))

        assert len(
            configs_with_name) > 0, f"Could not find a custodian with name {custodian_name}"

        config = configs_with_name[0]

        # Instantiate the client
        key = custodian_name if custodian_name in CUSTODIAN_NAME_TO_CLIENT_ADAPTER else "default"
        client_class = CUSTODIAN_NAME_TO_CLIENT_ADAPTER[key]["client"]
        adapter_class = CUSTODIAN_NAME_TO_CLIENT_ADAPTER[key]["adapter"]

        if config.refreshTokenUrl is not None:
            client = client_class(config.apiBaseUrl, token, config.refreshTokenUrl)
        else:
            client = client_class(config.apiBaseUrl, token)

        # Adapt the client into the class Custodian
        return adapter_class(client)



from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from pydantic.generics import GenericModel

T = TypeVar("T")


class JsonRpcRequest(GenericModel, Generic[T]):
    jsonrpc = "2.0"
    method: str
    params: Optional[T]
    id: Optional[Union[str, int]]




import unittest
from unittest import mock

from mmisdk.adapters.bitgo_adapter import BitgoAdapter
from mmisdk.bitgo.bitgo_client import BitgoClient
from mmisdk.mocks.mock_bitgo_requests_get import mock_bitgo_requests_get
from mmisdk.mocks.mock_bitgo_requests_post import mock_bitgo_requests_post
from mmisdk.mocks.mock_legacy_tx_params import MOCK_LEGACY_TX_PARAMS


class BitgoAdapterTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        client = BitgoClient("http://some-url", "some-refresh-token")
        cls.custodian = BitgoAdapter(client)

    def test_should_be_defined(self):
        self.assertIsNotNone(self.custodian)

    @mock.patch('requests.get', side_effect=mock_bitgo_requests_get)
    def test_should_get_transaction(self, mock_get):
        transaction = self.custodian.get_transaction("95851701-e573-497a-9df8-fbca8bbf80e9")
        self.assertEqual(transaction.id, '95851701-e573-497a-9df8-fbca8bbf80e9')
        self.assertEqual(transaction.type, '1')
        self.assertEqual(transaction.from_, '0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75')
        self.assertEqual(transaction.to, '0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506')
        self.assertEqual(transaction.value, '1000000000000000000')
        self.assertEqual(transaction.gas, '260849')
        self.assertEqual(transaction.gasPrice, None)
        self.assertEqual(transaction.maxPriorityFeePerGas, None)
        self.assertEqual(transaction.maxFeePerGas, None)
        self.assertEqual(transaction.nonce, None)
        self.assertEqual(transaction.data, "0x7ff36ab500000000000000000000000000000000000000000000000006d5dbf25e950f2f0000000000000000000000000000000000000000000000000000000000000080000000000000000000000000b22505ee2aac7d52d4a218f5877cdbae3bbeec7500000000000000000000000000000000000000000000000000000000624263dd0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000b4fbf271143f4fbf7b91a5ded31805e42b2208d60000000000000000000000001f9840a85d5af5bf1d1762f925bdaddc4201f984")
        self.assertEqual(transaction.hash, "0x2b81204f61c7736cb4b9cf89b842d44081b476f35e521fb745d8e212c02b77b4")
        self.assertEqual(transaction.status.finished, False)
        self.assertEqual(transaction.status.signed, False)
        self.assertEqual(transaction.status.success, False)
        self.assertEqual(transaction.status.displayText, "Created")
        self.assertEqual(transaction.status.reason, "Unknown")

    def test_get_transactions_should_raise(self):
        with self.assertRaises(NotImplementedError):
            self.custodian.get_transactions("some-id")

    @mock.patch('requests.post', side_effect=mock_bitgo_requests_post)
    def test_should_create_transaction(self, mock_post):
        extra_params = {
            "walletId": "someWalletId",
        }
        transaction = self.custodian.create_transaction(MOCK_LEGACY_TX_PARAMS, extra_params)
        self.assertEqual(transaction.id, '6ca20474-f1e6-4f1f-8123-56ff097a06cd')
        self.assertEqual(transaction.type, '1')
        self.assertEqual(transaction.from_, '0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7')
        self.assertEqual(transaction.to, '0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75')
        self.assertEqual(transaction.value, '999000000')
        self.assertEqual(transaction.gas, '300000')
        self.assertEqual(transaction.gasPrice, None)
        self.assertEqual(transaction.maxPriorityFeePerGas, None)
        self.assertEqual(transaction.maxFeePerGas, None)
        self.assertEqual(transaction.nonce, None)
        self.assertEqual(transaction.data, "0x1111111111111111")
        self.assertEqual(transaction.hash, None)
        self.assertEqual(transaction.status.finished, False)
        self.assertEqual(transaction.status.signed, False)
        self.assertEqual(transaction.status.success, False)
        self.assertEqual(transaction.status.displayText, "Created")
        self.assertEqual(transaction.status.reason, "Unknown")


if __name__ == "__main__":
    unittest.main()




from typing import Optional

from mmisdk.common.base_model import BaseModel
from mmisdk.common.hex_string import HexString


class JsonRpcCreateTransactionMetadata(BaseModel):
    chainId: HexString
    originUrl: Optional[str]  # The web page/dapp where the transaction originated
    transactionCategory: Optional[str]  # The category of transaction, as best can be determined by the wallet
    note: Optional[str]  # A note to be attached to the transaction which can be specified by the user



from typing import Optional

from mmisdk.common.dec_string import DecString
from mmisdk.common.hex_string import HexString


def hexlify(dec_string: Optional[DecString]) -> Optional[HexString]:
    """
    Converts a decimal string to its hexadecimal representation. Support None input.
    """
    return hex(int(dec_string)) if dec_string is not None else None



import unittest
from unittest.mock import patch

from mmisdk.adapters.bitgo_adapter import BitgoAdapter
from mmisdk.adapters.cactus_adapter import CactusAdapter
from mmisdk.adapters.jsonrpc_adapter import JsonRpcAdapter
from mmisdk.adapters.qredo_adapter import QredoAdapter
from mmisdk.factory.custodian_factory import CustodianFactory
from mmisdk.mocks.mock_mmiconfig_requests_get import mock_mmiconfig_requests_get


class CustodianFactoryTest(unittest.TestCase):

    @patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_be_defined(self, mock_get):
        factory = CustodianFactory()
        self.assertIsNotNone(factory)

    @patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_fetch_configs(self, mock_requests):
        factory = CustodianFactory()
        self.assertIsNotNone(factory.custodians_config)
        self.assertGreater(len(factory.custodians_config), 0)

    @patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_get_supported_custodians(self, mock_requests):
        factory = CustodianFactory()
        supported_custodians = factory.get_supported_custodians()
        self.assertEqual(8, len(supported_custodians))
        self.assertIn("qredo", supported_custodians)
        self.assertIn("qredo-dev", supported_custodians)
        self.assertIn("cactus", supported_custodians)
        self.assertIn("cactus-dev", supported_custodians)
        self.assertIn("bitgo", supported_custodians)
        self.assertIn("bitgo-test", supported_custodians)
        self.assertIn("saturn", supported_custodians)

    @patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_create_custodian(self, mock_requests):
        factory = CustodianFactory()
        custodian = factory.create_for("qredo", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, QredoAdapter)
        custodian = factory.create_for("qredo-dev", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, QredoAdapter)
        custodian = factory.create_for("cactus", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, CactusAdapter)
        custodian = factory.create_for("cactus-dev", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, CactusAdapter)
        custodian = factory.create_for("bitgo", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, BitgoAdapter)
        custodian = factory.create_for("bitgo-test", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, BitgoAdapter)
        custodian = factory.create_for("saturn", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, JsonRpcAdapter)

    @patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_fail_unsupported_custodian(self, mock_requests):
        factory = CustodianFactory()
        with self.assertRaises(AssertionError):
            factory.create_for("whatever", "refresh_token")


if __name__ == "__main__":
    unittest.main()




from typing import Optional

from mmisdk.common.base_model import BaseModel


class JsonRpcTransactionStatus(BaseModel):
    finished: bool  # Whether the transaction has finised, e.g. it is completed, failed or cancelled
    signed: bool  # Whether the transaction has been signed
    submitted: Optional[bool]  # Whether the transaction was submitted to the blockchain
    success: Optional[bool]  # Whether the transaction was successful, i.e. it was included in a block and not reverted
    displayText: str  # Short text to display to the user
    reason: Optional[str]  # The reason for the transaction status




import unittest
from unittest import mock

from mmisdk.mmiconfig.mmiconfig_client import MMIConfigClient
from mmisdk.mocks.mock_mmiconfig_requests_get import mock_mmiconfig_requests_get


class CustodianConfigClientTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.client = MMIConfigClient()

    def test_should_be_defined(self):
        self.assertIsNotNone(self.client)

    @mock.patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_get_mmi_config(self, mock_get):
        mmi_config = self.client._MMIConfigClient__get_mmi_config("mock-mmi-config-api-url")
        self.assertIsNotNone(mmi_config)
        self.assertEqual(len(mmi_config.custodians), 1)
        self.assertEqual(mmi_config.custodians[0].refreshTokenUrl, "https://hi.ho")

    @mock.patch('requests.get', side_effect=mock_mmiconfig_requests_get)
    def test_should_get_merged_custodian_configs(self, mock_get):
        """ There are  6 legacy """
        custodian_configs = self.client.get_merged_custodian_configs()
        self.assertIsNotNone(custodian_configs)
        self.assertEqual(len(custodian_configs), 8)  # 6 legacy custodian + 2 prod + 1 dev - 1 override (qredo)


if __name__ == "__main__":
    unittest.main()



from mmisdk.mocks.mock_bitgo_create_tx_response import MOCK_BITGO_CREATE_TX_RESPONSE
from mmisdk.mocks.mock_response import MockResponse


def mock_bitgo_requests_post(*args, **kwargs):

    # Mock the request that creates a transaction
    if args[0].__contains__('/tx/build'):
        return MockResponse(MOCK_BITGO_CREATE_TX_RESPONSE, 201)

    return MockResponse(None, 404)



MOCK_MMICONFIG_PROD = {
    "portfolio": None,
    "custodians": [
        {
            "refreshTokenUrl": "https://hey.ho",
            "name": "qredo",
            "apiBaseUrl": "https://foo.bar",
        },
        {
            "refreshTokenUrl": "https://some.url",
            "name": "somename",
            "apiBaseUrl": "https://another.one",
        },
    ]
}



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_CACTUS_DEV"]

# Create the custodian, using the factory
custodian = factory.create_for("cactus-dev", token)

# Get all transactions
transactions = custodian.get_transactions(42)

print(transactions)
# [
#     Transaction(id='VURKJPZ2JVD888888000277', type='1', from_='0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29', to=None, value=None, gas='180549', gasPrice='1000000007', maxPriorityFeePerGas='1000000008', maxFeePerGas='1400000012', nonce='0', data=None, hash='0x27a8aa70a501c20c7614aac084add4482482b51447804ff23275b1fb21e63927', status=TransactionStatus(finished=True, submitted=True, signed=True, success=False, displayText='Completed', reason='Unknown')),
#     Transaction(id='VHV3LIBLVND888888000431', type='1', from_='0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29', to=None, value=None, gas='1500000', gasPrice='1000000008', maxPriorityFeePerGas='1000000000', maxFeePerGas='3500000012', nonce='4', data=None, hash='0xc062e8e89fa6447b59fc981fa6be352d012bd8bf92bb6833626bcb34c89ce455', status=TransactionStatus(finished=True, submitted=True, signed=True, success=False, displayText='Completed', reason='Unknown')),
#     Transaction(id='HJRW4ZQNRHD888888000441', type='1', from_='0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29', to=None, value=None, gas='201060', gasPrice='1000000008', maxPriorityFeePerGas='1000000000', maxFeePerGas='3500000012', nonce='5', data=None, hash='0xf706b11023c6a443d6b3fffb350aea88909e1bccafa836addb0081b99052fa91', status=TransactionStatus(finished=True, submitted=True, signed=True, success=False, displayText='Completed', reason='Unknown')),
#     ...
# ]



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_SATURN_DEV"]

# Create the custodian, using the factory
custodian = factory.create_for("saturn-dev", token)

# Build tx details
tx_params = {
    "from": "0xb2c77973279baaaf48c295145802695631d50c01",
    "to": "0x57f36031E223FabC1DaF93B401eD9F4F1Acc6904",
    "type": "2",
    "value": "1",  # in Wei
    "gas": "21000",
    "maxPriorityFeePerGas": "1500000014",
    "maxFeePerGas": "1500000000",
    "data": "0xa9059cbb000000000000000000000000099dbbd63f749a9802526b277b8f3aaa3c5e81b0000000000000000000000000000000000000000000000000000000011ee39a80"
}
jsonrpc_extra_params = {
    "chainId": "0x4",
    "originUrl": "https://www.example.com",
    "transactionCategory": "transfer",
    "note": "This is a note to trader"
}

# Create the tx from details and send it to the custodian
transaction = custodian.create_transaction(tx_params, jsonrpc_extra_params)

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='9fbbd8d1-ef8e-425a-8743-b1008ea22152'
# type='2'
# from_='0xb2c77973279baaaf48c295145802695631d50c01'
# to='0x57f36031E223FabC1DaF93B401eD9F4F1Acc6904'
# value='1'
# gas='21000'
# gasPrice=None
# maxPriorityFeePerGas='1500000014'
# maxFeePerGas='1500000000'
# nonce=None
# data='0xa9059cbb000000000000000000000000099dbbd63f749a9802526b277b8f3aaa3c5e81b0000000000000000000000000000000000000000000000000000000011ee39a80'
# hash=None
# status=TransactionStatus(finished=False, signed=False, submitted=False, success=False, displayText='Created', reason=None)



from typing import Optional

from pydantic import AnyHttpUrl
from pydantic import HttpUrl

from mmisdk.common.base_model import BaseModel


class CustodianConfig(BaseModel):
    name: str
    apiBaseUrl: Optional[AnyHttpUrl]
    refreshTokenUrl: Optional[AnyHttpUrl]



import unittest
from unittest import mock

from mmisdk.adapters.cactus_adapter import CactusAdapter
from mmisdk.cactus.cactus_client import CactusClient
from mmisdk.mocks.mock_cactus_requests_get import mock_cactus_requests_get
from mmisdk.mocks.mock_cactus_requests_post import mock_cactus_requests_post
from mmisdk.mocks.mock_legacy_tx_params import MOCK_LEGACY_TX_PARAMS


class CactusAdapterTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        client = CactusClient("http://some-url", "some-refresh-token")
        cls.custodian = CactusAdapter(client)

    def test_should_be_defined(self):
        self.assertIsNotNone(self.custodian)

    @mock.patch('requests.get', side_effect=mock_cactus_requests_get)
    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_get_transaction(self, mock_get, mock_post):
        transaction = self.custodian.get_transaction("VURKJPZ2JVD888888000277")
        self.assertEqual(transaction.id, 'VURKJPZ2JVD888888000277')

    @mock.patch('requests.get', side_effect=mock_cactus_requests_get)
    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_get_transactions(self, mock_get, mock_post):
        transactions = self.custodian.get_transactions(42)
        self.assertEqual(transactions[0].id, "VURKJPZ2JVD888888000277")
        self.assertEqual(transactions[1].id, "VHV3LIBLVND888888000431")

    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_create_transaction(self, mock_post):
        extra_params = {
            "chainId": "1",
        }
        transaction = self.custodian.create_transaction(MOCK_LEGACY_TX_PARAMS, extra_params)
        self.assertEqual(transaction.id, 'VHV3LIBLVND888888000431')


if __name__ == "__main__":
    unittest.main()



MOCK_CACTUS_TRANSACTION_DETAIL_0 = {
    "transactionStatus": "completed",
    "transactionHash": "0x27a8aa70a501c20c7614aac084add4482482b51447804ff23275b1fb21e63927",
    "custodian_transactionId": "VURKJPZ2JVD888888000277",
    "gasPrice": "1000000007",
    "maxFeePerGas": "1400000012",
    "maxPriorityFeePerGas": "1000000008",
    "gasLimit": "180549",
    "nonce": "0",
    "from": "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29",
    "signature": None,
}



from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress


class QredoNewTransaction(BaseModel):
    from_: EthereumAddress = Field(None, alias="from")
    to: EthereumAddress
    value: DecString
    gasPrice: Optional[DecString]
    maxPriorityFeePerGas:  Optional[DecString]
    maxFeePerGas:  Optional[DecString]
    gasLimit: DecString
    data: str
    chainID: DecString



from mmisdk.mocks.mock_cactus_transaction_detail_0 import MOCK_CACTUS_TRANSACTION_DETAIL_0
from mmisdk.mocks.mock_cactus_transaction_detail_1 import MOCK_CACTUS_TRANSACTION_DETAIL_1
from mmisdk.mocks.mock_response import MockResponse


def mock_cactus_requests_get(*args, **kwargs):
    """ This method will be used by the mock to replace requests.get"""
    # Mock the request that gets all transactions
    if args[0].__contains__('/transactions'):
        return MockResponse([MOCK_CACTUS_TRANSACTION_DETAIL_0, MOCK_CACTUS_TRANSACTION_DETAIL_1], 200)

    return MockResponse(None, 404)



from typing import List

from mmisdk.adapters.cactus_create_tx_extra_params import CactusCreateTxExtraParams
from mmisdk.cactus.cactus_client import CactusClient
from mmisdk.cactus.cactus_get_transactions_params import CactusGetTransactionsParams
from mmisdk.cactus.cactus_transaction_detail import CactusTransactionDetail
from mmisdk.cactus.cactus_tx_params import CactusTXParams
from mmisdk.common.custodian import Custodian
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction_params import EIP1559TxParams
from mmisdk.common.transaction_params import LegacyTxParams
from mmisdk.utils.map_transaction_status import map_transaction_status


class CactusAdapter(Custodian):

    def __init__(self, client: CactusClient) -> None:
        self.client = client

    def get_transaction(self, transaction_id, chain_id=1) -> Transaction:
        params = CactusGetTransactionsParams(
            chainId=chain_id,
            transactionId=transaction_id,
        )
        cactus_transaction_details = self.client.get_transactions(params)

        result = cactus_transaction_details[0]
        return self.__map_cactus_transaction_to_transaction(result)

    def get_transactions(self, chain_id=1) -> List[Transaction]:
        params = CactusGetTransactionsParams(
            chainId=chain_id,
        )
        cactus_transaction_details = self.client.get_transactions(params)

        return list(map(lambda cactus_tx: self.__map_cactus_transaction_to_transaction(cactus_tx), cactus_transaction_details))

    def create_transaction(self, tx_params, extra_params) -> Transaction:
        # Parse and type check params
        type_ = tx_params["type"] if 'type' in tx_params else 1
        tx_params_parsed = EIP1559TxParams(**tx_params) if type_ == 2 else LegacyTxParams(**tx_params)
        extra_params_parsed = CactusCreateTxExtraParams(**extra_params)

        gasPrice = tx_params_parsed.gasPrice if type_ == 1 else None
        maxPriorityFeePerGas = tx_params_parsed.maxPriorityFeePerGas if type_ == 2 else None
        maxFeePerGas = tx_params_parsed.maxFeePerGas if type_ == 2 else None

        cactus_tx_params = CactusTXParams(
            from_=tx_params_parsed.from_,
            to=tx_params_parsed.to,
            gasLimit=tx_params_parsed.gas,
            value=tx_params_parsed.value,
            data=tx_params_parsed.data,
            gasPrice=gasPrice,
            maxPriorityFeePerGas=maxPriorityFeePerGas,
            maxFeePerGas=maxFeePerGas
        )
        cactus_transaction_detail = self.client.create_transaction(extra_params_parsed.chainId, cactus_tx_params)
        return self.__map_cactus_transaction_to_transaction(cactus_transaction_detail)

    def __map_cactus_transaction_to_transaction(self, cactus_transaction_detail: CactusTransactionDetail) -> Transaction:
        return Transaction(
            id=cactus_transaction_detail.custodian_transactionId,
            type="1" if cactus_transaction_detail.gasPrice is not None else "2",
            from_=cactus_transaction_detail.from_,
            to=None,
            value=None,  # TODO Why no field "value" on the response from Cactus API?
            gas=cactus_transaction_detail.gasLimit,
            gasPrice=cactus_transaction_detail.gasPrice,
            maxPriorityFeePerGas=cactus_transaction_detail.maxPriorityFeePerGas,
            maxFeePerGas=cactus_transaction_detail.maxFeePerGas,
            nonce=cactus_transaction_detail.nonce if cactus_transaction_detail.nonce != "" else None,
            data=None,  # TODO Why no field "data" on the response from Cactus API?
            hash=cactus_transaction_detail.transactionHash,
            status=map_transaction_status(cactus_transaction_detail.transactionStatus, "Unknown"),
        )



from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

supported_custodians = factory.get_supported_custodians()
print(supported_custodians)

# [ "qredo", "qredo-dev", "cactus", ... ]




import unittest

from mmisdk.utils.decify import decify


class DecifyTest(unittest.TestCase):
    def test_decify_should_convert_hex_string_to_decimal_string(self):
        self.assertEqual(decify('0x0'), "0")
        self.assertEqual(decify('0x1'), "1")
        self.assertEqual(decify('0x2a'), "42")
        self.assertEqual(decify('0x55593c0'), "89494464")

    def test_decify_should_support_longs(self):
        self.assertEqual(decify('0x13426172c74d822b878fe800000000'), "100000000000000000000000000000000000")

    def test_decify_should_return_none_when_passed_none(self):
        self.assertEqual(decify(None), None)


if __name__ == "__main__":
    unittest.main()




from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class JsonRpcCreateTransactionParams(BaseModel):
    type: Optional[HexString]
    from_: EthereumAddress = Field(None, alias="from")  # In Python, "from" is a reserved keyword
    to: EthereumAddress
    gas: HexString  # Gas limit
    value: HexString
    data: Optional[HexString]
    maxPriorityFeePerGas: Optional[HexString]  # Maximum fee per gas the sender is willing to pay miners in wei
    # The maximum total fee per gas the sender is willing to pay (includes the network / base fee and miner / priority fee) in wei
    maxFeePerGas: Optional[HexString]
    # The maximum total fee per gas the sender is willing to pay (includes the network / base fee and miner / priority fee) in wei
    gasPrice: Optional[HexString]



from mmisdk.mocks.mock_cactus_transaction_detail_1 import MOCK_CACTUS_TRANSACTION_DETAIL_1
from mmisdk.mocks.mock_response import MockResponse


def mock_cactus_requests_post(*args, **kwargs):
    """ This method will be used by the mock to replace requests.post"""

    # Mock the request that creates the access token
    if args[0].__contains__('/tokens'):
        return MockResponse({"jwt": "some-access-token"}, 200)

    # Mock the request that creates a transaction
    if args[0].__contains__('/transactions'):
        return MockResponse(MOCK_CACTUS_TRANSACTION_DETAIL_1, 200)

    return MockResponse(None, 404)



"""
Override configs for legacy custodians.
The MMI configuration API doesn't provide the actual configs for legacy custodians,
so we deffine overrides in the list.
"""

from mmisdk.mmiconfig.mmi_config import MMIConfig

overrides = {
    "custodians": [
        {
            "name": "qredo",
            "apiBaseUrl": "https://api.qredo.network",
            "refreshTokenUrl": None,
        },
        {
            "name": "qredo-dev",
            "apiBaseUrl": "https://7ba211-api.qredo.net",
            "refreshTokenUrl": None,
        },
        {
            "name": "cactus",
            "apiBaseUrl": "https://api.mycactus.com/custody/v1/mmi-api",
            "refreshTokenUrl": None,
        },
        {
            "name": "cactus-dev",
            "apiBaseUrl": "https://api.mycactus.dev/custody/v1/mmi-api",
            "refreshTokenUrl": None,
        },
        {
            "name": "bitgo",
            "apiBaseUrl": "https://app.bitgo.com/defi/v1/mmi",
            "refreshTokenUrl": None,
        },
        {
            "name": "bitgo-test",
            "apiBaseUrl": "https://app.bitgo-test.com/defi/v1/mmi",
            "refreshTokenUrl": None,
        }
    ]
}

MMICONFIG_LEGACY_OVERRIDE = MMIConfig(**overrides)



import requests

from mmisdk.common.abstract_client import AbstractClient
from mmisdk.qredo.qredo_new_transaction import QredoNewTransaction
from mmisdk.qredo.qredo_transaction_info import QredoTransactionInfo
from cachetools import TTLCache


class QredoClient(AbstractClient):

    def __init__(self, api_url, token):
        self.api_url = api_url
        self.token = token
        self.cache = TTLCache(maxsize=1, ttl=86400)  # Default TTL. The cache get re-created later with actual token expiry

    def __create_access_token(self):
        """Internal method that creates an access token, necessary to authenticate other calls against Qredo's API.

        Returns:
            The access token as a string.
        """
        url = self.api_url+'/connect/token'
        querystring = {"grant_type": "refresh_token",
                       "refresh_token": self.token}
        payload = ""

        response = requests.post(url, data=payload, params=querystring)
        if (response.status_code != 200):
            raise requests.HTTPError(
                f"Couldn't create the access token, {response.text}")

        response_json = response.json()
        access_token = response_json["access_token"]
        expires_in = response_json["expires_in"]
        return access_token, expires_in

    def __create_access_token_cached(self):
        """Lazily fetch the access token, looking up in the cache before. If the access is not in the cache, we fetch it then store it.
        Note that we have to dynamically re-recreate the cache after fetching the token, because the TTL we want to use is extracted from the token's expires_in.
        """
        if 'access_token' not in self.cache:
            access_token, expires_in = self.__create_access_token()
            self.cache = TTLCache(maxsize=1, ttl=expires_in)
            self.cache['access_token'] = access_token

        return self.cache['access_token']

    def __get_headers(self):
        """Returns the HTTP header to use in requests to the custodian"""
        access_token = self.__create_access_token_cached()
        return {
            "Content-Type": "application/json",
            "Authorization": 'Bearer ' + access_token
        }

    def get_transaction(self, tx_id: str) -> QredoTransactionInfo:
        url = self.api_url + '/connect/transaction/' + tx_id
        payload = ""
        headers = self.__get_headers()

        response = requests.get(url, data=payload, headers=headers)
        if (response.status_code != 200):
            raise requests.HTTPError(
                f"Couldn't get the transaction, {response.text}")

        # Parse response
        response_json = response.json()
        response_json["from_"] = response.json()["from"]

        return QredoTransactionInfo(**response_json)

    def create_transaction(self, new_transaction: QredoNewTransaction) -> QredoTransactionInfo:
        url = self.api_url + "/connect/transaction"
        headers = self.__get_headers()

        payload = new_transaction.dict()
        payload["from"] = payload["from_"]
        response = requests.post(url, json=payload, headers=headers)

        if (response.status_code != 200):
            raise requests.HTTPError(
                f"Couldn't create the transaction, {response.text}")

        # Parse response
        response_json = response.json()
        return QredoTransactionInfo(**response_json)



def map_transaction_status(status: str, reason: str):
    """
    Maps custodians' transaction statuses to JSON-RPC statuses, as documented in
    https://consensys.gitlab.io/codefi/products/mmi/mmi-docs/faq/transactions/?h=status#how-do-transaction-statuses-work

    Args:
        status: str - The transaction status string, in the custodian's nomenclature
        reason: Optional[str] - The reason for the signed message status, for instance "Signed message was rejected by the user", or "Signed message was cancelled by the user"
    Returns:
        A dictionary with the JSON-RPC status fields.
    """
    if status == 'created':
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Created',
            "reason": reason,
        }
    if status == 'pending':  # TODO TBC
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Pending',
            "reason": reason,
        }
    if status == 'authorized':  # TODO TBC
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Authorized',
            "reason": reason,
        }
    if status == 'approved':  # TODO TBC
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Created',
            "reason": reason,
        }
    if status == 'signed':
        return {
            "finished": False,
            "submitted": False,
            "signed": True,
            "success": False,
            "displayText": 'Signed',
            "reason": reason,
        }
    if status == 'submitted':
        return {
            "finished": False,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Submitted',
            "reason": reason,
        }
    if status == 'mined':
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": True,
            "displayText": 'Mined',
            "reason": reason,
        }
    if status == 'completed':  # TODO TBC
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Completed',
            "reason": reason,
        }
    if status == 'aborted':
        return {
            "finished": True,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Aborted',
            "reason": reason,
        }
    if status == 'rejected':  # TODO TBC
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Rejected',
            "reason": reason,
        }
    if status == 'failed':
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Failed',
            "reason": reason,
        }
    if status == 'overriden':  # TODO TBC
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Overriden',
            "reason": reason,
        }
    else:
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Unknown',
            "reason": reason,
        }



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from an environment variable, or anywhere else
token = os.environ["MMISDK_TOKEN_BITGO_TEST"]

# Create the custodian, using the factory
custodian = factory.create_for("bitgo-test", token)

# Get the transaction
transaction = custodian.get_transaction("0a4a244b-dbd4-4a4a-bc3d-0cf8f2f5f618")

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='0a4a244b-dbd4-4a4a-bc3d-0cf8f2f5f618'
# type='1'
# from_='0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7'
# to='0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75'
# value='999000000'
# gas='300000'
# gasPrice=None
# maxPriorityFeePerGas=None
# maxFeePerGas=None
# nonce=None
# data='0x1111111111111111'
# hash=None
# status=TransactionStatus(finished=False, submitted=False, signed=False, success=False, displayText='Created', reason='Unknown')



from mmisdk.mocks.mock_bitgo_get_tx_by_custodian_id import MOCK_BITGO_GET_TX_BY_CUSTODIAN_ID
from mmisdk.mocks.mock_response import MockResponse


def mock_bitgo_requests_get(*args, **kwargs):
    """ This method will be used by the mock to replace requests.get"""

    if args[0].__contains__('INVALID-CUSTODIAN-TX-ID'):
        return MockResponse({
            "error": "Did not find match for tx id: INVALID-CUSTODIAN-TX-ID",
            "errorName": "Error",
            "reqId": "unk-rktoh37uhl63sm8i1vfe",
            "context": {
                "_meta": {
                    "reqId": "unk-rktoh37uhl63sm8i1vfe"
                }
            }
        }, 404)

    if args[0].__contains__('/eth/wallets/transactions/'):
        return MockResponse(MOCK_BITGO_GET_TX_BY_CUSTODIAN_ID, 200)

    return MockResponse(None, 404)





class AbstractClient:
    """Abstract class that each custodian client inherits from."""

    def __init__(self, api_url, token):
        """
        Args:
            api_url: The custodian's API base URL.
            token: A token provided to you by the custodian. Depending on the custodian, it could be either a refresh token
            or an access token. Regardless of which, the library takes care of the authentication flow.
        """
        self.api_url = api_url
        self.token = token



import os

from mmisdk.factory.custodian_factory import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_SATURN_DEV"]

custodian = factory.create_for("saturn-dev", token)

# Get the transaction
transaction = custodian.get_transaction("94f1316b-98af-4736-8116-6faf72eab60c")

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='5377c945-f756-4c25-902d-c0d3b9cd179b'
# type='2'
# from_='0x033e270c08c3f297f99660f958d5f615207c1adf'
# to='0x7603A62b21A85f5cD02baE3389F35F1AcBaB0Ab2'
# value='0'
# gas='33997'
# gasPrice=None
# maxPriorityFeePerGas='1500000000'
# maxFeePerGas='19245046202'
# nonce='91'
# data='0x97c5ed1e000000000000000000000000033e270c08c3f297f99660f958d5f615207c1adf0000000000000000000000000000000000000000000000000de0b6b3a7640000'
# hash='0xe8073973a8ba8ba8d28b8a60e705c89f2facca54262fbc88a4f26f373ac29c81'
# status=TransactionStatus(finished=True, signed=True, submitted=True, success=True, displayText='Mined', reason=None)



from typing import Dict
from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString

# Using "Optional" on all fields as we're missing specific API docs for BitGo


class BitgoTransaction(BaseModel):
    transactionStatus: Optional[str]
    custodianTransactionId: Optional[str]
    from_: Optional[EthereumAddress] = Field(None, alias="from")
    to: Optional[EthereumAddress]
    coin: Optional[str]
    value: Optional[DecString]
    gasLimit: Optional[DecString]
    userId: Optional[str]
    createdTime: Optional[str]
    data: Optional[HexString]
    decodedData: Optional[Dict]
    transactionHash: Optional[HexString]




import unittest
from unittest import mock

import requests

from mmisdk.jsonrpc.jsonrpc_client import JsonRpcClient
from mmisdk.jsonrpc.jsonrpc_create_transaction_metadata import JsonRpcCreateTransactionMetadata
from mmisdk.jsonrpc.jsonrpc_create_transaction_params import JsonRpcCreateTransactionParams
from mmisdk.mocks.mock_jsonrpc_requests_post import mock_jsonrpc_requests_post


class JsonRpcClientTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.client = JsonRpcClient("http://some-json-rpc-url", "some-json-rpc-refresh-token", "http://some-json-rpc-refresh-token-url")

    def test_should_be_defined(self):
        self.assertIsNotNone(self.client)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_create_access_token(self, mock_post):
        access_token, expires_in = self.client._JsonRpcClient__create_access_token()
        self.assertEqual(access_token, "some-json-rpc-access-token")
        self.assertEqual(expires_in, 86400)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_get_headers(self, mock_post):
        headers = self.client._JsonRpcClient__get_headers()
        self.assertEqual(headers, {
            "Authorization": 'Bearer some-json-rpc-access-token'
        })

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_get_transaction_by_id(self, mock_post):
        get_transaction_result = self.client.get_transaction_by_id("ef8cb7af-1a00-4687-9f82-1f1c82fbef54")

        self.assertIsNotNone(get_transaction_result)
        self.assertEqual(get_transaction_result.id, "b903b4d6-bc70-435a-a289-24b04af0d78f")
        self.assertEqual(get_transaction_result.type, "0x2")
        self.assertEqual(get_transaction_result.from_, "0x367dacaf26cc1381f31911676ba40088938aa677")
        self.assertEqual(get_transaction_result.to, "0x7603A62b21A85f5cD02baE3389F35F1AcBaB0Ab2")
        self.assertEqual(get_transaction_result.value, "0x0")
        self.assertEqual(get_transaction_result.gas, "0x84c1")
        self.assertEqual(get_transaction_result.gasPrice, None)
        self.assertEqual(get_transaction_result.maxPriorityFeePerGas, "0x59682f00")
        self.assertEqual(get_transaction_result.maxFeePerGas, "0x596a67af")
        self.assertEqual(get_transaction_result.nonce, "0xa4f")
        self.assertEqual(get_transaction_result.data,
                         "0x97c5ed1e000000000000000000000000367dacaf26cc1381f31911676ba40088938aa6770000000000000000000000000000000000000000000000000de0b6b3a7640000")
        self.assertEqual(get_transaction_result.hash, "0x4b5981d66479ab1c7253de0b9e040dc2c74e38a5a005f1f253c9eb67cda31241")
        self.assertEqual(get_transaction_result.status.finished, True)
        self.assertEqual(get_transaction_result.status.submitted, True)
        self.assertEqual(get_transaction_result.status.signed, True)
        self.assertEqual(get_transaction_result.status.success, True)
        self.assertEqual(get_transaction_result.status.displayText, "Mined")
        self.assertEqual(get_transaction_result.status.reason, None)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_get_transaction_by_id_with_bad_id_should_raise(self, mock_post):
        with self.assertRaises(requests.HTTPError):
            self.client.get_transaction_by_id("bad-id")

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_create_transaction(self, mock_post):
        jsonrpc_create_tx_params = JsonRpcCreateTransactionParams(
            type="0x2",
            from_="0x367dacaf26cc1381f31911676ba40088938aa677",
            to="0x57f36031E223FabC1DaF93B401eD9F4F1Acc6904",
            gas="0x5208",
            value="0x1",
            maxPriorityFeePerGas="0x59682f0e",
            maxFeePerGas="0x59682f0e"
        )
        jsonrpc_create_tx_metadata = JsonRpcCreateTransactionMetadata(
            chainId="0x4",
            originUrl="https://www.example.com",
            note="This is a note to trader"
        )
        tx_id = self.client.create_transaction(jsonrpc_create_tx_params, jsonrpc_create_tx_metadata)
        self.assertEqual(tx_id, "b903b4d6-bc70-435a-a289-24b04af0d78f")

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_create_transaction_with_bad_from_should_raise(self, mock_post):
        with self.assertRaises(requests.HTTPError):
            jsonrpc_create_tx_params = JsonRpcCreateTransactionParams(
                type="0x2",
                from_="0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",  # Bad address
                to="0x57f36031E223FabC1DaF93B401eD9F4F1Acc6904",
                gas="0x5208",
                value="0x1",
                maxPriorityFeePerGas="0x59682f0e",
                maxFeePerGas="0x59682f0e"
            )
            jsonrpc_create_tx_metadata = JsonRpcCreateTransactionMetadata(
                chainId="0x4",
                originUrl="https://www.example.com",
                note="This is a note to trader"
            )
            self.client.create_transaction(jsonrpc_create_tx_params, jsonrpc_create_tx_metadata)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test___create_access_token_cached_should_call_once(self, mock_post):
        access_token_1 = self.client._JsonRpcClient__create_access_token_cached()
        access_token_2 = self.client._JsonRpcClient__create_access_token_cached()
        access_token_3 = self.client._JsonRpcClient__create_access_token_cached()
        access_token_4 = self.client._JsonRpcClient__create_access_token_cached()
        access_token_5 = self.client._JsonRpcClient__create_access_token_cached()
        mock_post.assert_called_once()
        self.assertEqual(access_token_1, access_token_2)
        self.assertEqual(access_token_2, access_token_3)
        self.assertEqual(access_token_3, access_token_4)
        self.assertEqual(access_token_4, access_token_5)


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class TransactionStatus(BaseModel):
    finished: bool  # Whether the transaction has finised, e.g. it is completed, failed or cancelled
    signed: bool  # Whether the transaction has been signed
    submitted: Optional[bool]  # Whether the transaction was submitted to the blockchain
    success: Optional[bool]  # Whether the transaction was successful, i.e. it was included in a block and not reverted
    displayText: str  # Short text to display to the user
    reason: Optional[str]  # The reason for the transaction status


class Transaction(BaseModel):
    id: str
    type: DecString
    from_: EthereumAddress = Field(None, alias="from")
    to: Optional[EthereumAddress]
    value: Optional[DecString]  # In Wei
    gas: Optional[DecString]  # Same as gasLimit
    gasPrice: Optional[DecString]
    maxPriorityFeePerGas: Optional[DecString]
    maxFeePerGas: Optional[DecString]
    nonce: Optional[DecString]
    data: Optional[HexString]
    hash: Optional[HexString]
    status: TransactionStatus



from pydantic import constr

"""
Pydantic field constraint, that matches only the empty string.
Typically combined with an other constraint, under a union like so:

Union[HexString, EmptyString]: Would match either a hexadecimal string, or the empty string.
"""

EmptyString = constr(regex=".?")



import unittest
from unittest import mock

import requests

from mmisdk.bitgo.bitgo_client import BitgoClient
from mmisdk.bitgo.bitgo_create_tx_params import BitgoCreateTxParams
from mmisdk.mocks.mock_bitgo_requests_get import mock_bitgo_requests_get
from mmisdk.mocks.mock_bitgo_requests_post import mock_bitgo_requests_post


class BitgoClientTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.client = BitgoClient("http://some-url", "some-bitgo-jwt-token")

    def test_should_be_defined(self):
        self.assertIsNotNone(self.client)

    def test_should_get_headers(self):
        headers = self.client._BitgoClient__get_headers()
        self.assertEqual(headers, {
            "Content-Type": "application/json",
            "Authorization": 'Bearer some-bitgo-jwt-token'
        })

    @mock.patch('requests.get', side_effect=mock_bitgo_requests_get)
    def test_should_get_transaction(self, mock_get):
        bitgo_list_response = self.client.get_transaction_by_custodian_tx_id("fdee94d9-5e91-447d-ba36-97f42a5c7dc6")

        self.assertIsNotNone(bitgo_list_response.data)
        self.assertEqual(len(bitgo_list_response.data), 1)
        self.assertIsNotNone(bitgo_list_response.meta)

        transaction = bitgo_list_response.data[0]
        self.assertEqual(transaction.transactionStatus, "created")
        self.assertEqual(transaction.custodianTransactionId, "95851701-e573-497a-9df8-fbca8bbf80e9")
        self.assertEqual(transaction.from_, "0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75")
        self.assertEqual(transaction.to, "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506")
        self.assertEqual(transaction.coin, "gteth")
        self.assertEqual(transaction.value, "1000000000000000000")
        self.assertEqual(transaction.gasLimit, "260849")
        self.assertEqual(transaction.userId, "5e822adc91a83a3f00975fa72b5a536d")
        self.assertEqual(transaction.createdTime, "2022-03-29T01:12:10.016Z")
        self.assertEqual(transaction.data, "0x7ff36ab500000000000000000000000000000000000000000000000006d5dbf25e950f2f0000000000000000000000000000000000000000000000000000000000000080000000000000000000000000b22505ee2aac7d52d4a218f5877cdbae3bbeec7500000000000000000000000000000000000000000000000000000000624263dd0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000b4fbf271143f4fbf7b91a5ded31805e42b2208d60000000000000000000000001f9840a85d5af5bf1d1762f925bdaddc4201f984")
        self.assertEqual(transaction.decodedData["protocolName"], "SushiswapV2Router")

    @mock.patch('requests.get', side_effect=mock_bitgo_requests_get)
    def test_should_raise_on_invalid_custodian_tx_id(self, mock_get):
        with self.assertRaises(requests.HTTPError):
            self.client.get_transaction_by_custodian_tx_id("INVALID-CUSTODIAN-TX-ID")

    @mock.patch('requests.post', side_effect=mock_bitgo_requests_post)
    def test_should_create_transaction(self, mock_post):
        wallet_id = "614a2a61df1332000686516a6dbee90a"
        bitgo_create_tx_params = BitgoCreateTxParams(txParams={
            "from_": "0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7",
            "to": "0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75",
            "value": "999000000",
            "gasPrice": "100",
            "gasLimit": "300000",
            "data": "0x1111111111111111"
        })
        bitgo_response = self.client.create_transaction(wallet_id, bitgo_create_tx_params)

        self.assertIsNotNone(bitgo_response.data)

        transaction = bitgo_response.data
        self.assertEqual(transaction.transactionStatus, "created")
        self.assertEqual(transaction.custodianTransactionId, "6ca20474-f1e6-4f1f-8123-56ff097a06cd")
        self.assertEqual(transaction.from_, "0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7")
        self.assertEqual(transaction.to, "0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75")
        self.assertEqual(transaction.coin, "gteth")
        self.assertEqual(transaction.gasLimit, "300000")
        self.assertEqual(transaction.value, "999000000")
        self.assertEqual(transaction.data, "0x1111111111111111")
        self.assertEqual(transaction.createdTime, "2022-09-22T08:19:55.843Z")


if __name__ == "__main__":
    unittest.main()



from typing import Any
from typing import List

from mmisdk.common.base_model import BaseModel
from mmisdk.mmiconfig.custodian_config import CustodianConfig


class MMIConfig(BaseModel):
    custodians: List[CustodianConfig]
    portfolio: Any




from typing import List
from typing import Union

import requests
from mmisdk.common.abstract_client import AbstractClient
from mmisdk.jsonrpc.jsonrpc_create_transaction_metadata import JsonRpcCreateTransactionMetadata
from mmisdk.jsonrpc.jsonrpc_create_transaction_params import JsonRpcCreateTransactionParams
from mmisdk.jsonrpc.jsonrpc_get_transaction_result import JsonRpcGetTransactionResult
from mmisdk.jsonrpc.jsonrpc_request import JsonRpcRequest
from mmisdk.jsonrpc.jsonrpc_response import JsonRpcResponse
from mmisdk.jsonrpc.jsonrpc_transaction_id import JsonRpcTransactionId
from cachetools import TTLCache


class JsonRpcClient(AbstractClient):

    def __init__(self, api_url, token, refresh_token_url):
        self.api_url = api_url
        self.token = token
        self.refresh_token_url = refresh_token_url
        self.cache = TTLCache(maxsize=1, ttl=86400)  # Default TTL. The cache get re-created later with actual token expiry
        self.use_cache = True  # Only used internally for comparing perf with/without caching

    def __create_access_token(self):
        """Internal method that creates an access token, necessary to authenticate other calls against the JSON-RPC custodian's API.

        Returns:
            The access token as a string.
        """
        payload = {"grant_type": "refresh_token",
                   "refresh_token": self.token}
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        response = requests.post(self.refresh_token_url,  data=payload, headers=headers)
        response_json = response.json()
        if 'access_token' not in response_json:
            raise requests.HTTPError(
                f"Couldn't create the access token, {response_json}")

        access_token = response_json["access_token"]
        expires_in = response_json["expires_in"]
        return access_token, expires_in

    def __create_access_token_cached(self):
        """Lazily fetch the access token, looking up in the cache before. If the access is not in the cache, we fetch it then store it.
        Note that we have to dynamically re-recreate the cache after fetching the token, because the TTL we want to use is extracted from the token's expires_in.
        """
        if not self.use_cache:
            access_token, _ = self.__create_access_token()
            return access_token

        if 'access_token' not in self.cache:
            access_token, expires_in = self.__create_access_token()
            self.cache = TTLCache(maxsize=1, ttl=expires_in)
            self.cache['access_token'] = access_token

        return self.cache['access_token']

    def __get_headers(self):
        """Internal method that creates the HTTP header to use for the calls against JSON-RPC custodian's API. It includes an authorization header that uses the access token.

        Returns:
            The HTTP header as a dictionary.
        """
        access_token = self.__create_access_token_cached()
        return {
            "Authorization": 'Bearer ' + access_token
        }

    def get_transaction_by_id(self, tx_id: str) -> JsonRpcGetTransactionResult:
        url = f"{self.api_url}/v1/json-rpc"
        headers = self.__get_headers()
        payload = JsonRpcRequest[List[str]](
            id=1,  # TODO
            method="custodian_getTransactionById",
            params=[tx_id]
        )
        payload_json = payload.dict()

        response = requests.post(url, json=payload_json, headers=headers)
        response_json = response.json()
        response_json_parsed = JsonRpcResponse[JsonRpcGetTransactionResult](**response_json)
        error = response_json_parsed.error
        result = response_json_parsed.result

        if error is not None:
            raise requests.HTTPError(
                f"Couldn't get the transaction. Request failed with code {error.code}: {error.message}")

        if result is None:
            raise requests.HTTPError(
                f"Coudn't find a transaction with passed id {tx_id}")

        return result

    def create_transaction(self, tx_params: JsonRpcCreateTransactionParams, tx_metadata: JsonRpcCreateTransactionMetadata) -> JsonRpcTransactionId:
        url = f"{self.api_url}/v1/json-rpc"
        headers = self.__get_headers()
        payload = JsonRpcRequest[List[Union[JsonRpcCreateTransactionParams, JsonRpcCreateTransactionMetadata]]](
            id=1,  # TODO
            method="custodian_createTransaction",
            params=[tx_params, tx_metadata],
        )

        # Exclude None fields, otherwise the JSON-RPC API raises Validation error
        payload_json = payload.dict(exclude_none=True, by_alias=True)

        response = requests.post(url, json=payload_json, headers=headers)
        response_json = response.json()
        response_json_parsed = JsonRpcResponse[JsonRpcTransactionId](**response_json)
        error = response_json_parsed.error

        if error is not None:
            raise requests.HTTPError(
                f"Couldn't create the transaction. Th request failed: {error}")

        # Parse and validate response
        return response_json["result"]



from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class CactusTXParams(BaseModel):
    """ Refer to ./custody-defi-api-4.yaml for more details """
    from_: EthereumAddress = Field(None, alias="from")
    to: EthereumAddress
    gasLimit: DecString
    value: DecString
    data: Optional[HexString]  # In practice it is indeed optional, contrary to what custody-defi-api-4.yaml says
    gasPrice: Optional[DecString]
    maxFeePerGas: Optional[DecString]
    maxPriorityFeePerGas: Optional[DecString]



import unittest
from unittest import mock

from mmisdk.adapters.qredo_adapter import QredoAdapter
from mmisdk.mocks.mock_legacy_tx_params import MOCK_LEGACY_TX_PARAMS
from mmisdk.mocks.mock_qredo_requests_get import mock_qredo_requests_get
from mmisdk.mocks.mock_qredo_requests_post import mock_qredo_requests_post
from mmisdk.qredo.qredo_client import QredoClient


class QredoAdapterTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        client = QredoClient("http://some-url", "some-refresh-token")
        cls.custodian = QredoAdapter(client)

    def test_should_be_defined(self):
        self.assertIsNotNone(self.custodian)

    def test_should_get_transactions(self):
        with self.assertRaises(NotImplementedError):
            self.custodian.get_transactions(1)

    @mock.patch('requests.get', side_effect=mock_qredo_requests_get)
    @mock.patch('requests.post', side_effect=mock_qredo_requests_post)
    def test_should_get_transaction(self, mock_get, mock_post):
        transaction = self.custodian.get_transaction("2ELuLA4HWIzPBQB1MSLmbOxtoB1")
        self.assertEqual(transaction.id, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')
        self.assertEqual(transaction.type, '1')
        self.assertEqual(transaction.status.finished, False)
        self.assertEqual(transaction.status.submitted, False)
        self.assertEqual(transaction.status.signed, False)
        self.assertEqual(transaction.status.success, False)
        self.assertEqual(transaction.status.displayText, "Authorized")
        self.assertEqual(transaction.status.reason, "Unknown")

    @mock.patch('requests.post', side_effect=mock_qredo_requests_post)
    def test_should_create_transaction(self, mock_post):
        extra_params = {
            "chainID": "1",
        }
        transaction = self.custodian.create_transaction(MOCK_LEGACY_TX_PARAMS, extra_params)
        self.assertEqual(transaction.id, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')


if __name__ == "__main__":
    unittest.main()



from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    """
    Configure globally the behaviour of pydantic.

    @see https://pydantic-docs.helpmanual.io/usage/model_config/#change-behaviour-globally
    """
    class Config:
        allow_population_by_field_name = True
        validate_assignment = True



from mmisdk.mocks.mock_mmiconfig_dev import MOCK_MMICONFIG_DEV
from mmisdk.mocks.mock_mmiconfig_prod import MOCK_MMICONFIG_PROD
from mmisdk.mocks.mock_response import MockResponse


def mock_mmiconfig_requests_get(*args, **kwargs):
    """ This method will be used by the mock to replace requests.get"""

    if args[0].__contains__('prod'):
        return MockResponse(MOCK_MMICONFIG_PROD, 200)

    return MockResponse(MOCK_MMICONFIG_DEV, 200)



from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from pydantic.generics import GenericModel

from mmisdk.jsonrpc.jsonrpc_error import JsonRpcError

T = TypeVar("T")


class JsonRpcResponse(GenericModel, Generic[T]):
    jsonrpc = "2.0"
    result: Optional[T]
    error: Optional[JsonRpcError]
    id: Optional[Union[str, int]]



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_QREDO_DEV"]

# Create the custodian, using the factory
custodian = factory.create_for("qredo-dev", token)

# Build tx details
tx_params = {
    "from": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "to": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "value": "100000000000000000",  # in Wei
    "gas": "21000",
    "gasPrice": "1000",
    # "data": "0xsomething",
    # "type": "2"
    # "maxPriorityFeePerGas": "12321321",
    # "maxFeePerGas": "12321321",
}
qredo_extra_params = {
    "chainID": "3",
}

# Create the tx from details and send it to the custodian
transaction = custodian.create_transaction(tx_params, qredo_extra_params)
print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='2EzDJkLVIjmH6LZQ2W1T4wPcTtK'
# type='1'
# from_='0x62468FD916bF27A3b76d3de2e5280e53078e13E1'
# to='0x62468FD916bF27A3b76d3de2e5280e53078e13E1'
# value='100000000000000000'
# gas='21000'
# gasPrice='1000'
# maxPriorityFeePerGas=None
# maxFeePerGas=None
# nonce='0'
# data=''
# hash=''
# status=TransactionStatus(finished=False, submitted=False, signed=False, success=False, displayText='Created', reason='Unknown')



import sys

if not 'sdist' in sys.argv:
    sys.exit(
        '\n ✋✋✋✋✋✋✋✋'
        '\n 🛑 The package `mmisdk` has been renamed `metamask-institutional.sdk`, and is no longer maintained under the name `mmisdk`.'
        '\n ✅ Use instead https://pypi.org/project/metamask-institutional.sdk.'
        '\n ✋✋✋✋✋✋✋✋')

from setuptools import setup


setup(
    name='mmisdk',
    version='0.5.1',
    description='The package `mmisdk` has been renamed `metamask-institutional.sdk`, and is no longer maintained under the name `mmisdk`. Use instead https://pypi.org/project/metamask-institutional.sdk.',
    author='Xavier Brochard',
    author_email='xavier.brochard@consensys.net',
    url='https://pypi.org/project/metamask-institutional.sdk/',
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3 :: Only',
        # 'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
        'Topic :: Office/Business :: Financial'
    ],
    project_urls={
        'Documentation': 'https://consensys.gitlab.io/codefi/products/mmi/mmi-sdk-py/sdk-python/',
        # 'Changelog': 'https://gitlab.com/ConsenSys/codefi/products/mmi/mmi-sdk-py/-/blob/main/CHANGELOG.md',
        # 'Issue Tracker': 'https://gitlab.com/ConsenSys/codefi/products/mmi/mmi-sdk-py/-/issues',
    },
    keywords='python sdk custodian interact get create transaction',
    python_requires='>=3.6',
    install_requires=[
        'pydantic>=1.10.1',
        'requests>=2.28.1',
        'cachetools>=5.2.0',
    ],
    extras_require={
        "dev": [
            "bump2version==1.0.1",
            "check-manifest==0.48",
            "pytest==7.1.3",
            "twine==4.0.1",
            "tox==3.26.0",
            "python-dotenv==0.21.0"
        ]
    },
    setup_requires=[
        'pytest-runner',
    ],
    # entry_points={
    #     'console_scripts': [
    #         'mmisdk = mmisdk.cli:main',
    #     ]
    # },
)



import os

from mmisdk.factory.custodian_factory import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_FPG_ALPHA"]
custodian = factory.create_for("fpg-alpha", token)

# Get the transaction
transaction = custodian.get_transaction("2ab09f15-3a7d-47fe-a8df-ba88af80f941")

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='2ab09f15-3a7d-47fe-a8df-ba88af80f941'
# type='2'
# from_='0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0'
# to='0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0'
# value='1'
# gas='21000'
# gasPrice=None
# maxPriorityFeePerGas='1500000014'
# maxFeePerGas='1500000014'
# nonce=None
# data=None
# hash=None
# status=TransactionStatus(finished=False, signed=False, submitted=True, success=False, displayText='FPG Transaction', reason='')




import unittest

from mmisdk.utils.hexlify import hexlify


class HexlifyTest(unittest.TestCase):
    def test_hexlify_should_convert_hex_string_to_decimal_string(self):
        self.assertEqual(hexlify('0'), "0x0")
        self.assertEqual(hexlify('1'), "0x1")
        self.assertEqual(hexlify('42'), "0x2a")
        self.assertEqual(hexlify('89494464'), "0x55593c0")

    def test_hexlify_should_support_longs(self):
        self.assertEqual(hexlify('100000000000000000000000000000000000'), "0x13426172c74d822b878fe800000000")

    def test_hexlify_should_return_none_when_passed_none(self):
        self.assertEqual(hexlify(None), None)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock

from mmisdk.cactus.cactus_client import CactusClient
from mmisdk.cactus.cactus_get_transactions_params import CactusGetTransactionsParams
from mmisdk.cactus.cactus_tx_params import CactusTXParams
from mmisdk.mocks.mock_cactus_requests_get import mock_cactus_requests_get
from mmisdk.mocks.mock_cactus_requests_post import mock_cactus_requests_post


class CactusClientTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.client = CactusClient("http://some-url", "some-refresh-token")

    def test_should_be_defined(self):
        self.assertIsNotNone(self.client)

    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_create_access_token(self, mock_post):
        access_token = self.client._CactusClient__create_access_token()
        self.assertEqual(access_token, "some-access-token")

    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_get_headers(self, mock_post):
        headers = self.client._CactusClient__get_headers()
        self.assertEqual(headers, {
            "Content-Type": "application/json",
            "Authorization": 'Bearer some-access-token'
        })

    @mock.patch('requests.get', side_effect=mock_cactus_requests_get)
    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_get_transaction(self, mock_get, mock_post):
        params = CactusGetTransactionsParams(chainId=42, transactionId="VURKJPZ2JVD888888000277")
        results = self.client.get_transactions(params)
        cactus_transaction_detail_1 = results[0]
        cactus_transaction_detail_2 = results[1]

        self.assertEqual(cactus_transaction_detail_1.transactionStatus, "completed")
        self.assertEqual(cactus_transaction_detail_1.transactionHash, "0x27a8aa70a501c20c7614aac084add4482482b51447804ff23275b1fb21e63927")
        self.assertEqual(cactus_transaction_detail_1.custodian_transactionId, "VURKJPZ2JVD888888000277")
        self.assertEqual(cactus_transaction_detail_1.gasPrice, "1000000007")
        self.assertEqual(cactus_transaction_detail_1.maxFeePerGas, "1400000012")
        self.assertEqual(cactus_transaction_detail_1.maxPriorityFeePerGas, "1000000008")
        self.assertEqual(cactus_transaction_detail_1.gasLimit, "180549")
        self.assertEqual(cactus_transaction_detail_1.nonce, "0")
        self.assertEqual(cactus_transaction_detail_1.from_, "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29")
        self.assertEqual(cactus_transaction_detail_1.signature, None)

        self.assertEqual(cactus_transaction_detail_2.transactionStatus, "completed")
        self.assertEqual(cactus_transaction_detail_2.transactionHash, "0xc062e8e89fa6447b59fc981fa6be352d012bd8bf92bb6833626bcb34c89ce455")
        self.assertEqual(cactus_transaction_detail_2.custodian_transactionId, "VHV3LIBLVND888888000431")
        self.assertEqual(cactus_transaction_detail_2.gasPrice, "1000000008")
        self.assertEqual(cactus_transaction_detail_2.maxFeePerGas, "3500000012")
        self.assertEqual(cactus_transaction_detail_2.maxPriorityFeePerGas, "1000000000")
        self.assertEqual(cactus_transaction_detail_2.gasLimit, "1500000")
        self.assertEqual(cactus_transaction_detail_2.nonce, "4")
        self.assertEqual(cactus_transaction_detail_2.from_, "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29")
        self.assertEqual(cactus_transaction_detail_2.signature, None)

    @mock.patch('requests.get', side_effect=mock_cactus_requests_get)
    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_get_transactions(self, mock_get, mock_post):
        cactus_get_transactions_params = CactusGetTransactionsParams(
            chainId=42,
        )
        transactions = self.client.get_transactions(cactus_get_transactions_params)
        tx0 = transactions[0]
        self.assertEqual(tx0.transactionStatus, "completed")
        self.assertEqual(tx0.transactionHash, "0x27a8aa70a501c20c7614aac084add4482482b51447804ff23275b1fb21e63927")
        self.assertEqual(tx0.custodian_transactionId, "VURKJPZ2JVD888888000277")
        self.assertEqual(tx0.gasPrice, "1000000007")
        self.assertEqual(tx0.maxFeePerGas, "1400000012")
        self.assertEqual(tx0.maxPriorityFeePerGas, "1000000008")
        self.assertEqual(tx0.gasLimit, "180549")
        self.assertEqual(tx0.nonce, "0")
        self.assertEqual(tx0.from_, "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29")
        self.assertEqual(tx0.signature, None)

        tx1 = transactions[1]
        self.assertEqual(tx1.transactionStatus, "completed")
        self.assertEqual(tx1.transactionHash, "0xc062e8e89fa6447b59fc981fa6be352d012bd8bf92bb6833626bcb34c89ce455")
        self.assertEqual(tx1.custodian_transactionId, "VHV3LIBLVND888888000431")
        self.assertEqual(tx1.gasPrice, "1000000008")
        self.assertEqual(tx1.maxFeePerGas, "3500000012")
        self.assertEqual(tx1.maxPriorityFeePerGas, "1000000000")
        self.assertEqual(tx1.gasLimit, "1500000")
        self.assertEqual(tx1.nonce, "4")
        self.assertEqual(tx1.from_, "0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29")
        self.assertEqual(tx1.signature, None)

    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test_should_create_transaction(self, mock_post):
        """We mock the methods requests.get and requests.post via the above decorators"""
        chain_id = 42
        cactus_tx_params = CactusTXParams(
            from_="0xA7f5D02E141c5DF8A987833E01E4f0922A91cF29",
            to="0xF4312f38f1139C2aa1c1dA54EF38F9ef1628dcB9",
            gasLimit="1500000",
            value="10000",
            data="0x0",
            gasPrice="1000000008",
            maxFeePerGas=None,
            maxPriorityFeePerGas=None
        )
        cactus_transaction_detail = self.client.create_transaction(chain_id, cactus_tx_params)
        self.assertEqual(cactus_transaction_detail.custodian_transactionId, 'VHV3LIBLVND888888000431')

    @mock.patch('requests.post', side_effect=mock_cactus_requests_post)
    def test___create_access_token_cached_should_call_once(self, mock_post):
        access_token_1 = self.client._CactusClient__create_access_token_cached()
        access_token_2 = self.client._CactusClient__create_access_token_cached()
        access_token_3 = self.client._CactusClient__create_access_token_cached()
        access_token_4 = self.client._CactusClient__create_access_token_cached()
        access_token_5 = self.client._CactusClient__create_access_token_cached()
        mock_post.assert_called_once()
        self.assertEqual(access_token_1, access_token_2)
        self.assertEqual(access_token_2, access_token_3)
        self.assertEqual(access_token_3, access_token_4)
        self.assertEqual(access_token_4, access_token_5)


if __name__ == "__main__":
    unittest.main()



from abc import abstractmethod
from typing import List
from typing import Union

from mmisdk.common.abstract_client import AbstractClient
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction_params import EIP1559TxParams
from mmisdk.common.transaction_params import LegacyTxParams


class Custodian(AbstractClient):
    @abstractmethod
    def get_transaction(self, transaction_id, chain_id=1) -> Transaction:
        """Gets the transaction of passed id, on passed chain id, that was submitted to the custodian.

         Args:
             transaction_id: The custodian-specific transaction id as a string.
             chain_id: (Optional) The id of the chain where the transaction happens. Default is 1.

         Returns:
             The transaction of passed id.
         """
        raise NotImplementedError()

    @abstractmethod
    def get_transactions(self, chain_id=1) -> List[Transaction]:
        """Gets all transactions on passed chain id that were submitted to the custodian. Scope might be limited to the scope of your refresh token.

        Args:
            chain_id: (Optional) The id of the chain where the transaction happens. Default is 1.

        Returns:
            The transaction of passed id.
        """
        raise NotImplementedError()

    @abstractmethod
    def create_transaction(self, tx_params: Union[LegacyTxParams, EIP1559TxParams], extra_params) -> Transaction:
        """
        Creates a transaction from the passed parameters a submits it to the custodian's API.

        Args:
            tx_params: The params to create the transaction with. Must map to either a LegacyTxParams or EIP1559TxParams.
            extra_params: Custodian-specific extra parameters. Refer to the types[custodian]_create_tx_extra_params.py for specific values.
        """
        raise NotImplementedError()



from typing import Optional

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class CactusGetTransactionsParams(BaseModel):
    """ Refer to ./custody-defi-api-4.yaml for more details """
    chainId: DecString
    from_: Optional[EthereumAddress] = Field(None, alias="from")
    transactionId: Optional[str]
    transactionHash: Optional[HexString]



import logging
from typing import Union

from mmisdk.adapters.bitgo_create_tx_extra_params import BitgoCreateTxExtraParams
from mmisdk.bitgo.bitgo_client import BitgoClient
from mmisdk.bitgo.bitgo_create_tx_params import BitgoCreateTxParams
from mmisdk.bitgo.bitgo_create_tx_params import BitgoCreateTxParamsInner
from mmisdk.bitgo.bitgo_transaction import BitgoTransaction
from mmisdk.common.custodian import Custodian
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction_params import EIP1559TxParams
from mmisdk.common.transaction_params import LegacyTxParams
from mmisdk.utils.map_transaction_status import map_transaction_status


class BitgoAdapter(Custodian):

    def __init__(self, client: BitgoClient) -> None:
        self.client = client

    def get_transaction(self, transaction_id, chain_id=1) -> Transaction:
        bitgo_list_response = self.client.get_transaction_by_custodian_tx_id(transaction_id)
        bitgo_transaction = bitgo_list_response.data[0]
        return self.__map_bitgo_transaction_to_transaction(bitgo_transaction)

    def create_transaction(self, tx_params: Union[LegacyTxParams, EIP1559TxParams], extra_params) -> Transaction:
        # Parse and type check params
        type_ = tx_params["type"] if 'type' in tx_params else 1
        tx_params_parsed = EIP1559TxParams(**tx_params) if type_ == 2 else LegacyTxParams(**tx_params)
        extra_params_parsed = BitgoCreateTxExtraParams(**extra_params)

        gasPrice = tx_params_parsed.gasPrice if type_ == 1 else None
        maxPriorityFeePerGas = tx_params_parsed.maxPriorityFeePerGas if type_ == 2 else None
        maxFeePerGas = tx_params_parsed.maxFeePerGas if type_ == 2 else None

        txParams = BitgoCreateTxParamsInner(
            from_=tx_params_parsed.from_,
            to=tx_params_parsed.to,
            value=tx_params_parsed.value,
            gasLimit=tx_params_parsed.gas,
            gasPrice=gasPrice,
            maxPriorityFeePerGas=maxPriorityFeePerGas,
            maxFeePerGas=maxFeePerGas,
            data=tx_params_parsed.data
        )

        bitgo_create_tx_params = BitgoCreateTxParams(txParams=txParams.dict(by_alias=True))

        bitgo_response = self.client.create_transaction(extra_params_parsed.walletId, bitgo_create_tx_params)
        bitgo_transaction = bitgo_response.data
        return self.__map_bitgo_transaction_to_transaction(bitgo_transaction)

    def __map_bitgo_transaction_to_transaction(self, bitgo_transaction: BitgoTransaction) -> Transaction:
        logging.warning(
            "Can't derive the transaction's type, because the BitGo's API returns incomplete data. Using type = '1' by default, but it might be wrong.")
        return Transaction(
            id=bitgo_transaction.custodianTransactionId,
            type="1",  # TODO We can't derive this since no field gasPrice/maxPriorityFeePerGas/maxFeePerGas on data returned by Bitgo's API
            from_=bitgo_transaction.from_,
            to=bitgo_transaction.to,
            value=bitgo_transaction.value,
            gas=bitgo_transaction.gasLimit,
            gasPrice=None,
            maxPriorityFeePerGas=None,
            maxFeePerGas=None,
            nonce=None,
            data=bitgo_transaction.data,
            hash=bitgo_transaction.transactionHash,
            status=map_transaction_status(bitgo_transaction.transactionStatus, "Unknown"),
        )



MOCK_QREDO_TRANSACTION_INFO = {
    "txID": "2ELuLA4HWIzPBQB1MSLmbOxtoB1",
    "txHash": "0x6b5dc3519b325be4e6f9512f8d3b1fdb7af22a996e3f6bb1a898a28ef05c1963",
    "status": "authorized",
    "timestamps": {
        "authorized": 1662387127,
        "created": 1662387126
    },
    "events": [{
        "id": "2ELuLAcfbNcFqCmYPHlaGlkIdQf",
        "timestamp": 1662387126,
        "status": "created",
        "message": ""
    },
        {
        "id": "2ELuLBuOP6tc1Pp09FSHKvfvOoZ",
        "timestamp": 1662387127,
        "status": "authorized",
        "message": ""
    }],
    "nonce": 0,
    "gasPrice": "1000",
    "gasLimit": "21000",
    "from": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "to":   "0x9999999999999999999999999999999999999999",
    "value": "1",
    "data": "",
    "rawTX": "4YCCA-iCUgiUYkaP2Ra_J6O3bT3i5SgOUweOE-EBgICAgA",
    "createdBy": "EFvSMt9uGTEsDCx22sYB8BPz1bCyxAZNPKDWpJVsLKM9",
    "accountID": "2ECrNK9dUlYIpdI4xVuRQ4Diuwq",
    "network": "",
    "chainID": "1"
}



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_FPG_ALPHA"]

# Create the custodian, using the factory
custodian = factory.create_for("fpg-alpha", token)

# Build tx details
tx_params = {
    "from": "0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0",
    "to": "0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0",
    "type": "2",
    "value": "1",  # in Wei
    "gas": "21000",
    "maxPriorityFeePerGas": "21000",
    "maxFeePerGas": "21000"
}
jsonrpc_extra_params = {
    "chainId": "0x1",  # FPG supports mainnet only
    "note": "This is a note to trader"
}

# Create the tx from details and send it to the custodian
transaction = custodian.create_transaction(tx_params, jsonrpc_extra_params)

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='2f0e06c0-b6d6-4c6b-b86f-518b1ed9f16d'
# type='2'
# from_='0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0'
# to='0x9BEe39bF784DDbEB0ae32BB1A7E7F927906C1da0'
# value='1'
# gas='21000'
# gasPrice=None
# maxPriorityFeePerGas='21000'
# maxFeePerGas='21000'
# nonce=None
# data=None
# hash=None
# status=TransactionStatus(finished=False, signed=False, submitted=True, success=False, displayText='FPG Transaction', reason='')



from mmisdk import CustodianFactory

# Use the factory to instantiate a custodian. It figures out itself which API and parameters to use.
factory = CustodianFactory()
custodian2 = factory.create_for("qredo-dev", "YOUR-REFRESH-TOKEN-QREDO-DEV")
custodian3 = factory.create_for("cactus-dev", "YOUR-REFRESH-TOKEN-CACTUS-DEV")



import requests

from mmisdk.bitgo.bitgo_create_tx_params import BitgoCreateTxParams
from mmisdk.bitgo.bitgo_list_response import BitgoListResponse
from mmisdk.bitgo.bitgo_response import BitgoResponse
from mmisdk.bitgo.bitgo_transaction import BitgoTransaction
from mmisdk.common.abstract_client import AbstractClient
from mmisdk.utils.inject_from_ import inject_from_on_all


class BitgoClient(AbstractClient):

    def __get_headers(self):
        """Internal method that creates the HTTP header to use for the calls against Cactus Custody's API. It includes an authorization header that uses the access token.

        Returns:
            The HTTP header as a dictionary.
        """
        return {
            "Content-Type": "application/json",
            "Authorization": 'Bearer ' + self.token
        }

    def get_transaction_by_custodian_tx_id(self, custodian_tx_id) -> BitgoListResponse[BitgoTransaction]:
        url = f"{self.api_url}/eth/wallets/transactions/{custodian_tx_id}"
        headers = self.__get_headers()

        response = requests.get(url, headers=headers)

        if response.status_code != 200:
            raise requests.HTTPError(
                f"Couldn't get the transaction. Request failed with status {response.status_code}: {response.json()}")

        # Parse response and build the result
        response_json = response.json()

        data_with_from_ = inject_from_on_all(response_json["data"])
        transactions = list(map(lambda json: BitgoTransaction(**json), data_with_from_))
        meta = response_json["_meta"]

        return BitgoListResponse(data=transactions, meta=meta)

    def create_transaction(self, wallet_id: str, tx_params: BitgoCreateTxParams) -> BitgoResponse[BitgoTransaction]:
        url = f"{self.api_url}/eth/wallet/{wallet_id}/tx/build"
        headers = self.__get_headers()
        payload = tx_params.dict(by_alias=True)
        response = requests.post(url, json=payload, headers=headers)

        if response.status_code != 201:
            raise requests.HTTPError(
                f"Couldn't create the transaction. Request failed with status {response.status_code}: {response.json()}")

        # Parse and return
        response_json = response.json()
        return BitgoResponse[BitgoTransaction](**response_json)



from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import Field

from mmisdk.common.base_model import BaseModel
from mmisdk.common.dec_string import DecString
from mmisdk.common.empty_string import EmptyString
from mmisdk.common.ethereum_address import EthereumAddress
from mmisdk.common.hex_string import HexString


class QredoTransactionEvent(BaseModel):
    id: str
    timestamp: int
    status: str
    message: str


class QredoTransactionStatus(str, Enum):
    PENDING = "pending"
    CREATED = "created"
    AUTHORIZED = "authorized"
    APPROVED = "approved"
    EXPIRED = "expired"
    CANCELLED = "cancelled"
    REJECTED = "rejected"
    SIGNED = "signed"
    SCHEDULED = "scheduled"
    PUSHED = "pushed"
    CONFIRMED = "confirmed"
    MINED = "mined"
    FAILED = "failed"


class QredoTransactionInfo(BaseModel):
    txID: str
    txHash: Union[HexString, EmptyString]
    status: QredoTransactionStatus
    nonce: int
    from_:  EthereumAddress = Field(None, alias="from")
    to: EthereumAddress
    value: DecString
    gasLimit: DecString
    maxPriorityFeePerGas: Optional[DecString]
    maxFeePerGas: Optional[DecString]
    gasPrice: Optional[DecString]
    data: Union[HexString, EmptyString]
    rawTX: str
    createdBy: str
    network: str
    chainID: DecString
    timestamps: Dict
    events: List[QredoTransactionEvent]





from typing import Union

from mmisdk.common.custodian import Custodian
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction import TransactionStatus
from mmisdk.common.transaction_params import EIP1559TxParams
from mmisdk.common.transaction_params import LegacyTxParams
from mmisdk.jsonrpc.jsonrpc_client import JsonRpcClient
from mmisdk.jsonrpc.jsonrpc_create_transaction_metadata import JsonRpcCreateTransactionMetadata
from mmisdk.jsonrpc.jsonrpc_create_transaction_params import JsonRpcCreateTransactionParams
from mmisdk.jsonrpc.jsonrpc_get_transaction_result import JsonRpcGetTransactionResult
from mmisdk.utils.decify import decify
from mmisdk.utils.hexlify import hexlify


class JsonRpcAdapter(Custodian):

    def __init__(self, client: JsonRpcClient) -> None:
        self.client = client

    def get_transaction(self, transaction_id, chain_id=1) -> Transaction:
        get_tx_result = self.client.get_transaction_by_id(transaction_id)
        return self.__map_jsonrpc_transaction_to_transaction(get_tx_result)

    def create_transaction(self, tx_params: Union[LegacyTxParams, EIP1559TxParams], extra_params) -> Transaction:
        # Parse and type check params
        type_ = tx_params["type"] if 'type' in tx_params else '1'
        tx_params_parsed = EIP1559TxParams(**tx_params) if type_ == '2' else LegacyTxParams(**tx_params)
        extra_params_parsed = JsonRpcCreateTransactionMetadata(**extra_params)

        gasPrice = tx_params_parsed.gasPrice if type_ == '1' else None
        maxPriorityFeePerGas = tx_params_parsed.maxPriorityFeePerGas if type_ == '2' else None
        maxFeePerGas = tx_params_parsed.maxFeePerGas if type_ == '2' else None

        jsonrpc_create_tx_params = JsonRpcCreateTransactionParams(
            type=hexlify(tx_params_parsed.type),
            from_=tx_params_parsed.from_,
            to=tx_params_parsed.to,
            gas=hexlify(tx_params_parsed.gas),
            value=hexlify(tx_params_parsed.value),
            data=tx_params_parsed.data,
            maxPriorityFeePerGas=hexlify(maxPriorityFeePerGas),
            maxFeePerGas=hexlify(maxFeePerGas),
            gasPrice=hexlify(gasPrice)

        )

        tx_id = self.client.create_transaction(jsonrpc_create_tx_params, extra_params_parsed)
        transaction = self.client.get_transaction_by_id(tx_id)
        return self.__map_jsonrpc_transaction_to_transaction(transaction)

    def __map_jsonrpc_transaction_to_transaction(self, jsonrpc_transaction: JsonRpcGetTransactionResult) -> Transaction:
        return Transaction(
            id=jsonrpc_transaction.id,
            type=decify(jsonrpc_transaction.type),  # Convert from hex to decimal
            from_=jsonrpc_transaction.from_,
            to=jsonrpc_transaction.to,
            value=decify(jsonrpc_transaction.value),
            gas=decify(jsonrpc_transaction.gas),
            gasPrice=decify(jsonrpc_transaction.gasPrice),
            maxPriorityFeePerGas=decify(jsonrpc_transaction.maxPriorityFeePerGas),
            maxFeePerGas=decify(jsonrpc_transaction.maxFeePerGas),
            nonce=decify(jsonrpc_transaction.nonce),
            data=jsonrpc_transaction.data,
            hash=jsonrpc_transaction.hash,
            status=TransactionStatus(**jsonrpc_transaction.status.dict(by_alias=True))
        )



import os

from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Grab your token from the environment, or anywhere else
token = os.environ["MMISDK_TOKEN_BITGO_TEST"]

# Create the custodian, using the factory
custodian = factory.create_for("bitgo-test", token)

# Build tx details
tx_params = {
    "from": "0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7",
    "to": "0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75",
    "value": "999000000",  # in Wei
    "gas": "123456",
    "gasPrice": "900000",
    "data": "0x00000"
}
bitgo_extra_params = {
    "walletId": "614a2a61df1332000686516a6dbee90a",
}

# Create the tx from details and send it to the custodian
transaction = custodian.create_transaction(tx_params, bitgo_extra_params)

print(type(transaction))
# <class 'mmisdk.common.transaction.Transaction'>

print(transaction)
# id='84236067-9962-49af-a2df-e635847289bd'
# type='1'
# from_='0xeddb59689e4ec3931b8d42d615c0d4a7a3a208e7'
# to='0xb22505ee2aac7d52d4a218f5877cdbae3bbeec75'
# value='999000000'
# gas='123456'
# gasPrice=None
# maxPriorityFeePerGas=None
# maxFeePerGas=None
# nonce=None
# data='0x00000'
# hash=None
# status=TransactionStatus(finished=False, submitted=False, signed=False, success=False, displayText='Created', reason='Unknown')




import unittest
from unittest import mock

from mmisdk.adapters.jsonrpc_adapter import JsonRpcAdapter
from mmisdk.common.transaction import Transaction
from mmisdk.jsonrpc.jsonrpc_client import JsonRpcClient
from mmisdk.mocks.mock_jsonrpc_requests_post import mock_jsonrpc_requests_post
from mmisdk.mocks.mock_legacy_tx_params import MOCK_LEGACY_TX_PARAMS


class JsonRpcAdapterTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        client = JsonRpcClient("http://some-url", "some-json-rpc-refresh-token", "http://some-json-rpc-refresh-token-url")
        cls.custodian = JsonRpcAdapter(client)

    def test_should_be_defined(self):
        self.assertIsNotNone(self.custodian)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_get_transaction(self, mock_get):
        transaction = self.custodian.get_transaction("b903b4d6-bc70-435a-a289-24b04af0d78f")
        self.__verify(transaction)

    @mock.patch('requests.post', side_effect=mock_jsonrpc_requests_post)
    def test_should_create_transaction(self, mock_post):
        extra_params = {
            "chainId": "0x4",
            "originUrl": "https://www.example.com",
            "note": "This is a note to trader"
        }
        transaction = self.custodian.create_transaction(MOCK_LEGACY_TX_PARAMS, extra_params)
        self.assertEqual(transaction.id, 'b903b4d6-bc70-435a-a289-24b04af0d78f')
        self.__verify(transaction)

    def __verify(self, transaction: Transaction) -> None:
        self.assertEqual(transaction.id, 'b903b4d6-bc70-435a-a289-24b04af0d78f')
        self.assertEqual(transaction.type, "2")
        self.assertEqual(transaction.from_, "0x367dacaf26cc1381f31911676ba40088938aa677")
        self.assertEqual(transaction.to, "0x7603A62b21A85f5cD02baE3389F35F1AcBaB0Ab2")
        self.assertEqual(transaction.value, "0")
        self.assertEqual(transaction.gas, "33985")
        self.assertEqual(transaction.gasPrice, None)
        self.assertEqual(transaction.maxPriorityFeePerGas, "1500000000")
        self.assertEqual(transaction.maxFeePerGas, "1500145583")
        self.assertEqual(transaction.nonce, "2639")
        self.assertEqual(
            transaction.data, "0x97c5ed1e000000000000000000000000367dacaf26cc1381f31911676ba40088938aa6770000000000000000000000000000000000000000000000000de0b6b3a7640000")
        self.assertEqual(transaction.hash, "0x4b5981d66479ab1c7253de0b9e040dc2c74e38a5a005f1f253c9eb67cda31241")
        self.assertEqual(transaction.status.finished, True)
        self.assertEqual(transaction.status.submitted, True)
        self.assertEqual(transaction.status.signed, True)
        self.assertEqual(transaction.status.success, True)
        self.assertEqual(transaction.status.displayText, "Mined")
        self.assertEqual(transaction.status.reason, None)


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from mmisdk.common.dec_string import DecString
from mmisdk.common.hex_string import HexString


def decify(hex_string: Optional[HexString]) -> Optional[DecString]:
    """
    Converts a hexadecimal string to its decimal representation. Support None input.
    """
    return str(int(hex_string, 16)) if hex_string is not None else None



from mmisdk.mocks.mock_jsonrpc_create_tx_response import MOCK_JSONRPC_CREATE_TX_RESPONSE
from mmisdk.mocks.mock_jsonrpc_empty_response import MOCK_JSONRPC_EMPTY_RESPONSE
from mmisdk.mocks.mock_jsonrpc_error_response import MOCK_JSONRPC_ERROR_RESPONSE
from mmisdk.mocks.mock_jsonrpc_get_tx_by_id import MOCK_JSONRPC_GET_TX_BY_ID
from mmisdk.mocks.mock_response import MockResponse


def mock_jsonrpc_requests_post(*args, **kwargs):
    """ This method will be used by the mock to replace requests.post"""

    json = kwargs["json"] if 'json' in kwargs else None
    method_field = json["method"] if json is not None and 'method' in json else None
    params_field = json["params"] if json is not None and 'params' in json else None
    from_field = params_field[0]["from"] if params_field is not None and len(params_field) > 0 and "from" in params_field[0] else None

    # Mock the request that creates the access token
    if args[0].__contains__('http://some-json-rpc-refresh-token-url'):
        return MockResponse({
            "access_token": "some-json-rpc-access-token",
            "refresh_token": "some-json-rpc-refresh-token",
            "token_type": "Bearer",
            "expires_in": 86400,
            "scope": ""
        }, 200)

    # Mock a response where no tx with passed id exists
    if params_field is not None and params_field.__contains__('bad-id'):
        return MockResponse(MOCK_JSONRPC_EMPTY_RESPONSE, 200)

    # Mock a response containing an error (this address is WETH's contract address, it's expected to fail)
    if from_field is not None and from_field.__contains__('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'):  # Bad address
        return MockResponse(MOCK_JSONRPC_ERROR_RESPONSE, 200)

    # Mock the request that gets a transaction by id
    if method_field is not None and method_field.__contains__('custodian_getTransactionById'):
        return MockResponse(MOCK_JSONRPC_GET_TX_BY_ID, 200)

    # Mock the request that creates a transaction by id
    if method_field is not None and method_field.__contains__('custodian_createTransaction'):
        return MockResponse(MOCK_JSONRPC_CREATE_TX_RESPONSE, 200)

    return MockResponse(None, 404)
